/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: hppa.h,v 1.8 1996/08/15 21:03:10 tiggr Exp $  */

#include <config/hpux/target.h>

#define APPLY_ARGS_REG_SIZE  (4 * sizeof (int) + 4 * sizeof (double))

#define APPLY_ARGS_START(SEL)  \
  do									\
    {									\
      int offset = 0;							\
      int reserved_size = (sizeof (void *) * (SEL)->out->num		\
			   + sizeof (tom_double) * (SEL)->in->num)

#define APPLY_ARGS_STACK_SIZE(SEL)  \
      reserved_size

/* Macro local to this target file to emit a value of some non-floating type
   to the ARGS being built.  */
#define APPLY_ARGS_EMIT_INT_VALUE(TYPE, VALUE)  \
  do									\
    {									\
      if (offset < 16)							\
	*(TYPE *) &args->reg[20 - offset] = (VALUE);			\
      else								\
	*(TYPE *) ((char *) (args->stack) + reserved_size - offset - 4)	\
	  = (VALUE);							\
      offset += sizeof (TYPE);						\
    } while (0)

/* Add a byte value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_BYTE(VALUE)  \
  APPLY_ARGS_EMIT_INT ((tom_int) (VALUE))

/* Add a char value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_CHAR(VALUE)  \
  APPLY_ARGS_EMIT_INT ((tom_int) (VALUE))

/* Add a int value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_INT(VALUE)  \
  APPLY_ARGS_EMIT_INT_VALUE (tom_int, VALUE)

/* Add a long value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_LONG(VALUE)  \
  do									\
    {									\
      offset = (offset + 7) & ~7;					\
      if (offset < 12)							\
	*(tom_long *) &args->reg[20 - offset - 4] = (VALUE);		\
      else								\
	*(tom_long *) ((char *) args->stack + reserved_size		\
		     - offset - 8) = (VALUE);				\
      offset += sizeof (tom_long);					\
    } while (0)

/* Add a float value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_FLOAT(VALUE)  \
  do									\
    {									\
      if (offset < 16)							\
	*(tom_double *) &args->reg[24 + (offset << 1)] = (VALUE);	\
      else								\
	*(tom_float *) ((char *) args->stack + reserved_size		\
			- offset - 4) = (VALUE);			\
      offset += sizeof (tom_float);					\
    } while (0)

/* Add a double value to the ARGS being built.  */
#define APPLY_ARGS_EMIT_DOUBLE(VALUE)  \
  do									\
    {									\
      if (offset < 12)							\
	*(tom_double *) &args->reg[24 + (offset << 1)] = (VALUE);	\
      else								\
	{								\
	  offset = (offset + 7) & ~7;					\
	  *(tom_double *) ((char *) args->stack + reserved_size		\
			   - offset - 8) = (VALUE);			\
	}								\
      offset += sizeof (tom_double);					\
    } while (0)

/* Add a selector to the ARGS being built.  */
#define APPLY_ARGS_EMIT_SELECTOR(VALUE)  \
  APPLY_ARGS_EMIT_POINTER (VALUE)

/* Add a reference to the ARGS being built.  */
#define APPLY_ARGS_EMIT_REFERENCE(VALUE)  \
  APPLY_ARGS_EMIT_POINTER (VALUE)

/* Add a pointer to the ARGS being built.  */
#define APPLY_ARGS_EMIT_POINTER(VALUE)  \
  APPLY_ARGS_EMIT_INT_VALUE (void *, VALUE)

#define APPLY_ARGS_COMPLETE()  \
      do							\
	{							\
	  if (offset < 16)					\
	    args->stack = 0;					\
	  else							\
	    args->stack = (void *) ((char *) args->stack	\
				    + reserved_size);	\
	} while (0)

#define APPLY_ARGS_ACTUAL_SIZE()  \
      (offset < 16 ? 0 : (offset + 7) & ~7)

#define APPLY_ARGS_END()  \
    } while (0)
