/* The Unicode utility.
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: uc.t,v 1.9 1996/08/07 15:31:06 tiggr Exp $  */

implementation class
uc: stdio
{
  <doc> The kind of thing we're supposed to do.  </doc>
  int to_do;

  <doc> The kind of thing we can do.  </doc>
  const CREATE_BITSET = 1;
  const CREATE_MAP = 2;
}

void
  usage
{
  [[err print ("usage: ", [Runtime program_name],
	       " [letter|digit|numeric|punctuation|map]")] nl];

  [Runtime exit 1]
}

(String, int)
  nextArgument (Array, int) (arguments, i)
{
  if (i == [arguments length] - 1)
    [self usage];

  = (arguments[++i], i)
}

void
  set_to_do int a
{
  if (to_do != 0)
    [self usage];
  to_do = a
}

int
  main Array arguments
{
  int i, n = [arguments length];
  MutableByteString t;
  selector predicate;

  in = [BufferedStream with in];

  for (i = 0; i < n; i++)
    {
      ByteString a = arguments[i];

      if (["map" equal a])
	[self set_to_do CREATE_MAP]
      else
	{
	  if (["letter" equal a])
	    predicate = selector ("o_isLetter")
	  else if (["digit" equal a])
	    predicate = selector ("o_isDigit")
	  else if (["numeric" equal a])
	    predicate = selector ("o_isNumeric")
	  else if (["punctuation" equal a])
	    predicate = selector ("o_isPunctuation")
	  else
	    [self usage];

	  [self set_to_do CREATE_BITSET]
	}
    };

  if (!to_do)
    [self usage]
  else if (to_do == CREATE_MAP)
    {
      MutableByteArray map = [MutableByteArray withCapacity 512];

      [map resize [map capacity]];

      do
	{
	  t = [in readLine];

	  /* Very dumb parser.  */
	  if (!!t && !![t length] && t[0] != '#')
	    {
	      Array fields = [t componentsSeparatedBy '\t' limit: 3];
	      int iso, uni;
	      boolean b;

	      if ([fields length] < 2)
		break;

	      (iso, b, ) = [fields[0] integerValue (0, -1) allowCBases: YES];
	      if (!b)
		break;

	      (uni, b, ) = [fields[1] integerValue (0, -1) allowCBases: YES];
	      if (!b)
		break;

	      if (iso < 0 || iso > 0xff || uni < 0 || uni > 0xffff)
		break;

	      map[2 * iso] = byte (uni / 256);
	      map[2 * iso + 1] = byte (uni % 256)
	    }
	} while (t != nil);

      if (t != nil)
	{
	  [[err print ("bad line: ", t)] nl];
	  [tom.Runtime exit 1]
	};

      [out writeBytes map]
    }
  else if (to_do == CREATE_BITSET)
    {
      MutableByteArray set = [MutableByteArray withCapacity 8192];

      [set resize [set capacity]];

      do
	{
	  t = [in readLine];

	  if (!!t && [t length] != 0)
	    {
	      UnicodeCharacter ch = [UnicodeCharacter with t];

	      if (ch == nil)
		{
		  [[err print ("error: ", t)] nl];
		  [Runtime exit 1]
		};

	      if ([ch perform predicate])
		{
		  char u = [ch unicode];
		  set[u / 8] = set[u / 8] | byte (1 << (u % 8))
		}
	    }
	} while (t != nil);

      [out writeBytes set]
    };

  = 0
}

end;

implementation instance uc end;
