<copyright> TOM parser generator bootstrap parser.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright (C) 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: GPParse.t,v 1.15 1996/09/07 20:02:21 tiggr Exp $</id>
    </copyright>

/******************** GPParseDefines ********************/

implementation class
GPParseDefines
{
  <doc> The string to output for a new line.  </doc>
  static MutableByteString indent;

  <doc> Repitition of elements of a rule.  </doc>
  const REPEAT_OPTIONAL = -1;
  const REPEAT_STAR = -2;
  const REPEAT_PLUS = -3;

  <doc> The token value (to be used in the look ahead set) for the empty
      production rule (i.e. epsilon).  </doc>
  const EPSILON = 65536;

  const IDENTIFIER = 65537;

  const NUMBER = 65538;

  /* Something between `{'... `}'.  */
  const EXPR = 65539;

  /* Something between `'' characters.  */
  const LITERAL = 65540;

  /* Real tokens.  */
  const TOKEN = 65541;
  const CLASS = 65542;

  /* Something between `%%', without the `%%'.  */
  const CODE = 65543;

  <doc> Iff YES, there were changes in the sets.  </doc>
  static boolean changes;
}

end;

implementation instance GPParseDefines end;

/******************** GPParse ********************/

implementation class
GPParse: Parser, GPParseDefines, ParserDefines

end;

implementation instance
GPParse
{
  <doc> The things declared to be a token, keyed on their name.  </doc>
  MutableDictionary tokens;

  <doc> The non-terminals, keyed on their name.  </doc>
  MutableDictionary non_terminals;

  // It isn't really public; it is available to our `friend' NonTerminal...
  // Wed Jun 12 15:34:39 1996, tiggr@cobra.es.ele.tue.nl
  <doc> The rules, keyed on the non-terminal's name.  </doc>
  public MutableDictionary rules;

  <doc> The name of the parser class.  </doc>
  String parser_name;

  <doc> The top rule.  </doc>
  Rule top;

  <doc> The next numeric value to be assigned to a token.  </doc>
  int next_token;
}

void
  addToken String ident
{
  if (tokens[ident] != nil)
    [[[self error] print ("redeclaration of token `", ident, '\'')] nl]
  else
    {
      Token t = [[Token alloc] init ident value next_token++];

      tokens[ident] = t;
      [tokens_by_number add ident]
    }
}

void
  dumpSets OutputStream str
{
  Enumerator e;
  boolean b;
  Rule r;

  e = [rules enumerator];
  while ({(b, r) = [e next]; b})
    {
      [str print ([r name], " first: ")];
      if (!![r first])
	[self describe str tokens [[r first] enumerator]]
      else
	[str print nil];

      [[str nl] print ([r name], " la: ")];
      if (!![r la])
	[self describe str tokens [[r la] enumerator]]
      else
	[str print nil];

      [[str nl] print ([r name], " follow: ")];
      if (!![r follow])
	[self describe str tokens [[r follow] enumerator]]
      else
	[str print nil];

      [str nl]
    }
}

OutputStream
  generateDefines OutputStream str
{
  Enumerator e;
  boolean b;
  String s;
  Token t;

  [str print ("implementation class ", parser_name, "Defines\n{")];

  e = [tokens keyEnumerator];
  while ({(b, s) = [e next]; b})
    [str print (indent, "const ", s, " = ", tokens[s], ';')];

  [[str print (indent, '}', indent, "end;")] nl];

  [[str print ("implementation instance ", parser_name, "Defines end;")] nl]
}

OutputStream
  generate OutputStream str
{
  Enumerator e;
  boolean b;
  String s;
  Rule r;

  current_parser = self;

  /* Walk the rules, checking for rules which are never used.  */
  e = [rules keyEnumerator];
  while ({(b, s) = [e next]; b})
    if (!non_terminals[s] && top != rules[s])
      [[err print ("warning: `", s, "' defined but not used")] nl];

  /* Walk the non-terminals, checking they have been defined.  */
  e = [non_terminals keyEnumerator];
  while ({(b, s) = [e next]; b})
    if (!rules[s])
      [[[self error] print ('`', s, "' used but not defined")] nl];

  if (!top)
    [[[self error] print "no top rule"] nl]
  else if (!parser_name)
    [[[self error] print "no parser class name"] nl]
  else if (!num_errors)
    {
      IntegerRangeSet eof = [IntegerRangeSet new];

      [eof add EOF];

      [top set_is_top YES];

      do
	{
	  changes = NO;
	  [Symbol set_current_run [Symbol current_run] + 1];
	  [top computeLA eof]
	} while (changes);

      if (trace)
	[self dumpSets err];

      if (!num_errors)
	{
	  [top checkLL1];

	  [str print ("<doc> Generated by ", [Runtime program_name],
		      " on ", [Date now], ".  </doc>\n")];

	  [self generateDefines str];

	  [str print ("implementation class ", parser_name,
		      ": Parser, ", parser_name, "Defines end;\n")];
	  [[str print ("implementation instance ", parser_name)] nl];

	  e = [rules enumerator];
	  while ({(b, r) = [e next]; b})
	    [r compile str];
	  [str nl];

	  [[str print ("end;")] nl]
	}
    };

  current_parser = nil;

  = str
}

<doc> Override designated initializer.  </doc>
redefine id
  init Lexer l
{
  rules = [MutableDictionary new];
  tokens = [MutableDictionary new];
  non_terminals = [MutableDictionary new];
  next_token = 65537;

  = [super init l]
}

<doc> This is the top rule.  </doc>
void
  grammar
{
  if (trace)
    [[err print ("entering grammar")] nl];

  indent = [MutableByteString new];
  [indent add '\n'];

  current_parser = self;

  [self nextToken];

  /* Initialize.  */
  tokens_by_number = [MutableObjectArray new];
  [tokens_by_number add "EPSILON"];

  while (token != EOF)
    {
      if (token == IDENTIFIER || token == TOKEN || token == CLASS)
	[self declaration]
      else
	[self fatalError]
    };

  current_parser = nil
}

private void
  declaration
{
  if (trace)
    [[err print ("entering declaration")] nl];

  if (token == IDENTIFIER)
    [self rule]
  else if (token == TOKEN)
    {
      [self nextToken];

      for (;;)
	{
	  if (token == IDENTIFIER)
	    {
	      [self addToken [lex semantics]];
	      [self nextToken]
	    }
	  else
	    [self fatalError];

	  if (token == ',')
	    [self nextToken]
	  else
	    break
	}
    }
  else if (token == CLASS)
    {
      [self nextToken];

      if (token == IDENTIFIER)
	{
	  if (!parser_name)
	    parser_name = [lex semantics]
	  else
	    [[[self error] print "parser class name already defined"] nl];
	  [self nextToken]
	}
      else
	[self fatalError]
    }
  else
    [self fatalError]
}

private void
  rule
{
  String expr;

  if (trace)
    [[err print ("entering rule")] nl];

  if (token != IDENTIFIER)
    [self fatalError];

  String nt = [lex semantics];
  if (tokens[nt] != nil)
    [[[self error] print ("token ", nt, " used as non-terminal")] nl]
  else if (rules[nt] != nil)
    [[[self error] print ("multiple rules for `", nt, '\'')] nl];
  [self nextToken];

  Array attributes;

  if (token == '(')
    attributes = [self attributes];

  if (token == EXPR)
    {
      expr = [lex semantics];
      [self nextToken]
    };

  [self term ':'];

  Rule r = [[Rule alloc] init nt start expr productions [self alternatives]
			 attributes attributes];
  if (top == nil)
    top = r;
  rules[nt] = r;

  [self term ';']
}

private Alternatives
  alternatives
{
  Alternatives r = [Alternatives new];

  if (trace)
    [[err print ("entering alternatives")] nl];

  [r add [self alternative]];

  while (token == '|')
    {
      [self nextToken];
      [r add [self alternative]]
    };

  = r
}

private Sequence
  alternative
{
  Sequence r = [Sequence new];

  if (trace)
    [[err print ("entering alternative")] nl];

  while (token != ';' && token != ']' && token != '|')
    if (token == EXPR)
      {
	[r add [[Element alloc] init nil attributes nil expr [lex semantics]]];
	[self nextToken]
      }
    else
      [r add [self repeats [self element]]];

  = r
}

private Element
  element
{
  String expr;
  Array a;

  if (trace)
    [[err print ("entering element")] nl];

  if (token == '[')
    {
      [self nextToken];

      = [[Element alloc] init [self alternatives] attributes nil expr nil];

      [self term ']']
    }
  else if (token == IDENTIFIER)
    {
      String name = [lex semantics];
      Symbol s = tokens[name];

      if (!s)
	{
	  s = non_terminals[name];
	  if (!s)
	    non_terminals[name]
	      = s = [[NonTerminal alloc] init name parser self]
	};

      [self nextToken];
      if (token == '(')
	a = [self actualAttributes];

      if (token == EXPR)
	{
	  expr = [lex semantics];
	  [self nextToken]
	};

      = [[Element alloc] init s attributes a expr expr]
    }
  else if (token == LITERAL)
    {
      Terminal lit = [lex semantics];

      [self nextToken];

      if (token == '(')
	{
	  a = [self actualAttributes];
	  String a_name = !a || [a length] != 1 ? nil : a[0];

	  if (!a_name)
	    [[[Symbol name [[self error] print "bad attributes to token `"]
		      token [lit intValue]] print '\''] nl]
	};

      if (token == EXPR)
	{
	  expr = [lex semantics];
	  [self nextToken]
	};

      = [[Element alloc] init lit attributes a expr expr]
    }
  else
    [self fatalError]
}

private Element
  repeats Element e
{
  if (trace)
    [[err print ("entering repeats")] nl];

  if (token == '*')
    {
      [e set_repeats REPEAT_STAR];
      [self nextToken]
    }
  else if (token == '+')
    {
      [e set_repeats REPEAT_PLUS];
      [self nextToken]
    }
  else if (token == '?')
    {
      [e set_repeats REPEAT_OPTIONAL];
      [self nextToken]
    };

  = e
}

private Array
  actualAttributes
{
  MutableArray attr = [MutableObjectArray new];

  if (trace)
    [[err print ("entering attributes")] nl];

  do
    {
      [self nextToken];

      if (token == EXPR || token == IDENTIFIER)
	{
	  [attr add [lex semantics]];
	  [self nextToken]
	}
      else
	[self fatalError]
    } while (token == ',');

  [self term ')'];

  = attr
}

private Array
  attributes
{
  MutableArray attr = [MutableObjectArray new];

  if (trace)
    [[err print ("entering attributes")] nl];

  [self term '('];
  [attr add [self attribute]];

  do
    if (token == ',')
      {
	[self nextToken];
	[attr add [self attribute]]
      }
    while (token != ')');

  [self nextToken];

  = attr
}

private Attribute
  attribute
{
  if (trace)
    [[err print ("entering attribute")] nl];

  if (token == '>' || token == '<')
    {
      boolean in_p = token == '>';
      String type, name;

      [self nextToken];
      type = [lex semantics];
      [self term IDENTIFIER];
      name = [lex semantics];
      [self term IDENTIFIER];

      = [[Attribute alloc] init (type, name) in in_p]
    }
  else
    [self fatalError]
}

end;
