<copyright> TOM parser generator grammar classes.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright (C) 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: grammar.t,v 1.18 1996/09/06 16:16:53 tiggr Exp $</id>
    </copyright>

/******************** Attribute ********************/

implementation class Attribute: State end;

implementation instance
Attribute
{
  public boolean in_not_out;
  public String type, name;
}

id
  init (String, String) (t, n)
    in boolean in_p
{
  (in_not_out, type, name) = (in_p, t, n);
  = self
}

end;

/******************** Alternatives ********************/

implementation class
Alternatives: Symbol, MutableObjectArray

end;

implementation instance
Alternatives

void
  checkLL1
{
  int i, j;

  for (i = 0; i < length; i++)
    {
      [self[i] checkLL1];

      for (j = i + 1; j < length; j++)
	{
	  IntegerRangeSet isec
	    = [[self[i] la] intersectionWith [self[j] la]];

	  if (![isec isEmpty])
	    {
	      OutputStream s = [[Parser current_parser] error self];

	      [s print ("overlap in alternatives ", i, " and ", j, ": ")];

	      [[[Parser current_parser]
		 describe s tokens [isec enumerator]] nl]
	    }
	}
    }
}

OutputStream
  compile OutputStream s
{
  if (length == 1)
    return [self[0] compile s];

  boolean first = YES;
  Symbol epsilon;
  int i;

  for (i = 0; i < length; i++)
    {
      Symbol sym = self[i];

      if ([[sym first] member EPSILON])
	if (epsilon)
	  [[[Parser error self] print "multiple EPSILON rules"] nl]
	else
	  epsilon = sym
      else
	{
	  if (first)
	    {
	      [s print indent];
	      first = NO
	    }
	  else
	    [s print (indent, "else ")];

	  [s print ("if ", [isa condition [[sym la] enumerator]])];
	  [indent add ' '];
	  [s print (indent, "{")];
	  [indent add ' '];

	  [sym compile s];

	  /* Avoid superfluous `;'.  */
	  [s print (indent, "void")];

	  [indent removeElementAt 1];
	  [s print (indent, "}")];
	  [indent removeElementAt 1]
	}
    };

  if (epsilon != nil)
    {
      if (!first)
	[s print (indent, "else")];

      [indent add ' '];
      [s print (indent, "{")];
      [indent add ' '];

      [epsilon compile s];

      /* Avoid superfluous `;'.  */
      [s print (indent, "void")];

      [indent removeElementAt 1];
      [s print (indent, "}")];
      [indent removeElementAt 1]
    }
  else
    [s print (indent, "else [self fatalError]")];

  = [s print ';']
}

IntegerRangeSet
  computeLA IntegerRangeSet f
{
  if (!follow)
    follow = f
  else if (run == current_run && [f equal follow])
    return la;
  run = current_run;

  IntegerRangeSet nf = [IntegerRangeSet new];
  if (!la)
    la = [IntegerRangeSet new];

  int i;
  for (i = 0; i < length; i++)
    {
      Symbol s = self[i];

      [la uniteWith [s computeLA follow]];
      [nf uniteWith [s first]]
    };

  if (first == nil || ![first equal nf])
    changes = YES;

  first = nf;

  = la
}

end;

/******************** Element ********************/

implementation class
Element: Symbol

end;

implementation instance
Element
{
  <doc> The symbol in this element in the enclosing rule.  </doc>
  Symbol symbol;

  <doc> Actual attributes, if any.  </doc>
  Array attributes;

  <doc> The expression (implementing this parser's semantics) following the
      {symbol}.  </doc>
  String expr;

  <doc> The kind of repitition of this symbol, or, if positive, how often.
      </doc>
  mutable int repeats;
}

void
  checkLL1
{
  if (symbol != nil)
    [symbol checkLL1]
}

OutputStream
  compile OutputStream s
{
  Enumerator e;
  int token;

  if (symbol)
    {
      String condition = [isa condition [[symbol la] enumerator]];

      if (repeats == REPEAT_OPTIONAL)
	[s print (indent, "if ", condition)]
      else if (repeats == REPEAT_STAR)
	[s print (indent, "while ", condition)]
      else if (repeats == REPEAT_PLUS)
	[s print (indent, "do")];

      [indent add ' '];
      [s print (indent, "{")];
      [indent add ' '];

      if (expr != nil && [symbol outputBefore attributes])
	[s print (indent, expr, ';')];

      if (!attributes)
	[symbol compile s]
      else
	[symbol compile s withAttributes attributes];

      if (expr != nil && ![symbol outputBefore attributes])
	[s print (indent, expr, ';')];

      /* Avoid superfluous `;'.  */
      [s print (indent, "void")];

      [indent removeElementAt 1];
      [s print (indent, "}")];
      [indent removeElementAt 1];

      if (repeats == REPEAT_PLUS)
	[s print (" while ", condition, ';')]
      else
	[s print ";"]
    }
  else if (expr)
    [s print (indent, expr, ';')];

  = s
}

IntegerRangeSet
  computeLA IntegerRangeSet f
{
  if (run == current_run && follow != nil && [follow equal f])
    return la;
  run = current_run;

  IntegerRangeSet s = [IntegerRangeSet new];
  follow = f;

  if (la != nil)
    [s uniteWith la];

  if (symbol == nil || repeats == REPEAT_OPTIONAL || repeats == REPEAT_STAR)
    [s uniteWith follow];

  if (symbol != nil)
    [s uniteWith [symbol computeLA follow]];

  boolean b = !la || ![la equal s];
  la = s;
  if (b && symbol != nil && (repeats == REPEAT_STAR || repeats == REPEAT_PLUS))
    [symbol computeLA la];

  if (!symbol)
    {
      first = [IntegerRangeSet new];
      [first add EPSILON]
    }
  else
    first = [symbol first];

  = la
}

id
        init Symbol s
  attributes Array a
        expr String e
{
  (symbol, attributes, expr) = (s, a, e);

  = [super init]
}

redefine boolean
  isTerminal
{
  = !symbol ? NO : [symbol isTerminal]
}

end;

/******************** NonTerminal ********************/

implementation class
NonTerminal: Symbol

end;

implementation instance
NonTerminal
{
  <doc> Our parser.  </doc>
  GPParse parser;

  <doc> Our name.  </doc>
  public String name;
}

void
  checkLL1
{
  [[parser rules][name] checkLL1]
}

OutputStream
	 compile OutputStream s
  withAttributes Array actual
{
  Array formal = [[parser rules][name] attributes];
  int n_attr = !actual ? 0 : [actual length];

  if (formal != nil && [formal length] != n_attr
      || !formal && n_attr != 0)
    [[[parser error self] print ("incorrect attributes to `", name, '\'')] nl]
  else
    {
      int n_out_attr, i;
      Attribute a;

      for (i = 0; i < n_attr; i++)
	{
	  a = formal[i];

	  if (![a in_not_out])
	    {
	      if (!n_out_attr++)
		[s print (indent, "(")]
	      else
		[s print ", "];
	      [s print {Any a1 = actual[i]}]
	    }
	};
      if (!n_out_attr)
	[s print (indent, "[self ", name)]
      else
	[s print (") = [self ", name)];

      if (n_attr != n_out_attr)
	{
	  boolean first = YES;

	  for (i = 0; i < n_attr; i++)
	    {
	      a = formal[i];

	      if ([a in_not_out])
		{
		  // CCC Likewise...
		  // Sat Jun 22 23:10:17 1996, tiggr@tricky.es.ele.tue.nl
		  [s print (String (first ? " (" : ", "), Any (actual[i]))];
		  first = NO		    
		}
	    };
	  [s print ')']
	};

      [s print "];"]
    };

  = s
}

IntegerRangeSet
  computeLA IntegerRangeSet f
{
  Rule r = [parser rules][name];

  follow = f;
  la = [r computeLA f];
  first = [r first];

  = la
}

<doc> Designated initializer </doc>
id
    init String n
  parser GPParse p
{
  (name, parser) = (n, p);

  = [super init]
}

end;

/******************** Rule ********************/

implementation class
Rule: State, Symbol

end;

implementation instance
Rule
{
  <doc> The name of the non-terminal.  </doc>
  public String name;

  <doc> The expression to be evaluated before the first grammar code.  </doc>
  String expr;

  <doc> The productions.  </doc>
  Alternatives rhs;

  <doc> The attributes.  </doc>
  public Array attributes;

  <doc> Whether we're the top rule.  </doc>
  mutable boolean is_top;

  <doc> Iff TRUE, we've checked for LL1'ness.  </doc>
  boolean checked_ll1;
}

void
  checkLL1
{
  if (!checked_ll1)
    {
      checked_ll1 = TRUE;
      [rhs checkLL1]
    }
}

OutputStream
  compile OutputStream s
{
  int i, n_attr, n_attr_out;
  Attribute a;

  /* Output the return type, aka the type of the synthesized attributes.  */
  if (attributes != nil)
    {
      n_attr = [attributes length];
      for (i = 0; i < n_attr; i++)
	{
	  a = attributes[i];

	  if (![a in_not_out])
	    {
	      if (!n_attr_out++)
		[s print (indent, '(')]
	      else
		[s print ", "];

	      [s print [a type]]
	    }
	};

      if (n_attr_out != 0)
	[s print ')']
    };

  if (!n_attr_out)
    [s print (indent, "void")];

  [s print (indent, "  ", name)];

  /* Output the arguments, aka the inherited attributes.  */
  if (n_attr != 0 && n_attr != n_attr_out)
    {
      boolean first = YES;

      for (i = 0; i < n_attr; i++)
	{
	  a = attributes[i];

	  if ([a in_not_out])
	    {
	      if (first)
		{
		  first = NO;
		  [s print " ("]
		}
	      else
		[s print ", "];

	      [s print [a type]]
	    }
	};

      for ({first = YES; i = 0}; i < n_attr; i++)
	{
	  a = attributes[i];

	  if ([a in_not_out])
	    {
	      if (first)
		{
		  first = NO;
		  [s print ") ("]
		}
	      else
		[s print ", "];

	      [s print [a name]]
	    }
	};

      [s print ')']
    };

  /* Start the body of the method.  */
  [s print (indent, "{")];
  [indent add ' '];

  if (is_top)
    {
      [s print (indent, "Parser previous_parser = current_parser;")];
      [s print (indent, "current_parser = self;")];
      [s print (indent, "unwind (current_parser = previous_parser)")];
      [s print (indent, "{")];
      [indent add ' '];
      [s print (indent, "[self nextToken];")]
    };

  if (n_attr_out != 0)
    for (i = 0; i < n_attr; i++)
      {
	a = attributes[i];

	if (![a in_not_out])
	  [s print (indent, [a type], ' ', [a name], ';')]
      };

  [s print (indent, "if (trace) [[err print \"entering rule: ",
	    name, "\"] nl];")];

  if (expr != nil)
    [s print (indent, expr, ';')];

  [rhs compile s];

  if (is_top)
    {
      [s print (indent, "if (token != EOF) [self fatalError]")];
      [indent removeElementAt 1];
      [s print (indent, "};")]
    };

  if (n_attr_out == 0)
    [s print (indent, "void")]
  else
    {
      boolean first = YES;

      [s print indent];
      for (i = 0; i < n_attr; i++)
	{
	  a = attributes[i];

	  if (![a in_not_out])
	    {
	      // CCC Without the cast to String, the compiler won't be
	      // able to type it...
	      // Sat Jun 22 21:47:48 1996, tiggr@tricky.es.ele.tue.nl
	      [s print (String (first ? {first = NO; "return ("} : ", "),
			[a name])]
	    }
	};
      [s print ')']
    };

  [indent removeElementAt 1];
  [s print (indent, "}")];

  = s
}

IntegerRangeSet
  computeLA IntegerRangeSet f
{
  if (!follow)
    follow = [IntegerRangeSet new]
  else if (run == current_run && [f equal follow])
    return la;
  run = current_run;

  [follow uniteWith f];

  la = [rhs computeLA follow];

  if (first == nil || ![first equal [rhs first]])
    {
      changes = YES;
      first = [rhs first]
    };

  = la
}

id
	 init String n
	start String e
  productions Alternatives p
   attributes Array a
{
  (name, expr, rhs, attributes) = (n, e, p, a);

  = [super init]
}

end;

/******************** Sequence ********************/

implementation class
Sequence: Symbol, MutableObjectArray

end;

implementation instance
Sequence

void
  checkLL1
{
  int i;

  for (i = 0; i < length; i++)
    [self[i] checkLL1]
}

OutputStream
  compile OutputStream s
{
  int i;

  for (i = 0; i < length; i++)
    s = [self[i] compile s];

  = s
}

IntegerRangeSet
  computeLA IntegerRangeSet f
{
  if (run == current_run && follow != nil && [follow equal f])
    return la;
  run = current_run;

  follow = la = f;

  int i;
  for (i = length - 1; i >= 0; i--)
    la = [self[i] computeLA la];

  IntegerRangeSet next = [IntegerRangeSet new];
  for (i = 0; i < length; i++)
    {
      IntegerRangeSet nf = [self[i] first];

      if (nf != nil)
	{
	  [next uniteWith nf];
	  if (![nf member EPSILON])
	    {
	      [next remove EPSILON];
	      break
	    }
	}
    };

  if (!first || ![first equal next])
    {
      changes = YES;
      first = next
    };

  = la
}

end;

/******************** Symbol ********************/

implementation class
Symbol: State, ParserDefines, GPParseDefines, FileElement
{
  <doc> The current run...  </doc>
  public mutable static int current_run;
}

<doc> Return a condition deciding whether the token matches the tokens
    returned by {e}.  </doc>
String
  condition Enumerator e
{
  MutableByteString s = [MutableByteString new];
  boolean b, first = YES;
  int token;

  while ({(b, token) = [e next]; b})
    {
      if (first)
	{
	  first = NO;
	  [s print '(']
	}
      else
	[s print " || "];

      [self name [s print "token == "] token token]
    };

  if (first)
    [s print "(1"];
  [s print ')'];

  = s
}

OutputStream
   name OutputStream s
  token int token
{
  = if (token >= ' ' && token <= '~')
      [s print ('\'', byte (token), '\'')]
    else if (token >= 65536)
      {
	String n = [[Parser current_parser] tokens_by_number][token - 65536];

	[s print n]
      }
    else
      [s print token]
}

end;

implementation instance
Symbol
{
  <doc> The first, look-ahead, and follow sets.  </doc>
  public IntegerRangeSet first, la, follow;

  <doc> The computation run in which we've last been updated.  </doc>
  int run;

  <doc> The line in the file at which we reside.  </doc>
  public int line;
}

<doc> Check LL(1)-ness of the receiving symbol.  </doc>
deferred void
  checkLL1;

<doc> Output code for parsing the receiving symbol.  </doc>
OutputStream
  compile OutputStream s
{
  = [self compile s withAttributes nil]
}

OutputStream
	 compile OutputStream s
  withAttributes Array actual
{
  [self shouldNotImplement cmd]
}

<doc> Compute and return the look-ahead set of the receiving symbol, given
    the follow set {follow}.  This computes all three significant sets of
    the receiving symbol.  </doc>
deferred IntegerRangeSet
  computeLA IntegerRangeSet follow;

<doc> Designated initializer.  </doc>
redefine id
  init
{
  line = [[[Parser current_parser] lex] current_line];
  = self
}

boolean
  outputBefore Array attributes
{
  return NO
}

boolean
  isTerminal
{
  = NO
}

end;

/******************** Terminal ********************/

implementation class
Terminal: Symbol, IntNumber

end;

implementation instance
Terminal

void
  checkLL1
{
  void
}

redefine boolean
  outputBefore Array attributes
{
  return attributes == nil
}

redefine OutputStream
	 compile OutputStream s
  withAttributes Array attributes
{
  if (attributes != nil)
    s = [s print ({Any a = attributes[0]}, " = [lex semantics];")];

  = [[isa name [s print (indent, "[self term ")] token value]
     print "];"]
}

IntegerRangeSet
  computeLA IntegerRangeSet follow
{
  if (!la)
    {
      la = first = [IntegerRangeSet new];
      [la add value]
    };

  = la
}

redefine boolean
  isTerminal
{
  = YES
}

end;

/******************** Token ********************/

implementation class
Token: Symbol, Terminal

end;

implementation instance
Token
{
  <doc> Our name.  </doc>
  public String name;
}

<doc> Designated initializer.  </doc>
id
   init String n
  value int v
{
  name = n;

  [super init v];

  = [super (Symbol) init]
}

end;
