<copyright> GP Parser skeleton.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: Parser.t,v 1.11 1996/09/05 14:58:30 tiggr Exp $</id>
    </copyright>

/******************** ParserDefines ********************/

implementation class
ParserDefines

end;

implementation instance ParserDefines end;

/******************** FileElement ********************/

implementation class
FileElement
end;

implementation instance
FileElement

<doc> Return the line in the file at which this element was declared (or
    whatever you call it).  </doc>
deferred int
  line;

end;

/******************** Parser ********************/

implementation class
Parser: State, stdio
{
  // This should not be 0, obviously.  However, 0 is what lex returns for
  // end of file...
  // Sun Jun  9 16:27:08 1996, tiggr@cobra.es.ele.tue.nl
  const EOF = 0;

  <doc> The path for all parsers to search.  </doc>
  static MutableIndexed path;

  <doc> The current parser.  </doc>
  static public Parser current_parser;

  <doc> The global number of errors counted.  </doc>
  int global_num_errors;

  <doc> Mapping from token number to token name.  </doc>
  MutableArray tokens_by_number;

  <doc> A default value for tracing, per parser class.  </doc>
  public mutable boolean trace_default;
}

void
  addToPath String directory
{
  if (!path)
    path = [MutableObjectArray new];

  [path add directory]
}

<doc> Return the stream to which to report an error by asking the
    {current_parser}.  </doc>
OutputStream
  error
{
  = current_parser ? [current_parser error]
      :
	{
	  global_num_errors++;
	  [stdio err]
	}
}

<doc> Return the stream to which to report an error for the {object},
    similar to {error}.  </doc>
OutputStream
  error FileElement object
{
  = (current_parser ? [current_parser error object]
     :
       {
	 global_num_errors++;
	 [[stdio err] print ([object line], ": ")]
       })
}

<doc> Return the {path}.  </doc>
Indexed
  path
{
  = path
}

end;

implementation instance
Parser
{
  <doc> The lexer from which we get the tokens.  </doc>
  public Lexer lex;

  <doc> The current token.  </doc>
  int token;

  <doc> The number of errors and warnings reported through this parser.  </doc>
  public int num_errors, num_warnings;

  <doc> Iff !0, we output all kinds of nice debugging information.  </doc>
  public mutable boolean trace;
}

<doc> Describe the {token} on the output stream {s}.  </doc>
OutputStream
  describe OutputStream s
     token int token
{
  = if (token >= ' ' && token <= '~')
      [s print ('\'', byte (token), '\'')]
    else if (token == 0)
      [s print "EOF"]
    else if (token >= 65536)
      if (!tokens_by_number)
	[s print token]
      else
	[s print {Any t = tokens_by_number[token - 65536]}]
    else
      [s print ("char(", token, ')')]
}

<doc> Describe the tokens from the enumerator {e} on the output stream {s}.
    </doc>
OutputStream
  describe OutputStream s
    tokens Enumerator e
{
  boolean first = YES;
  int i;

  while ({boolean b; (b, i) = [e next]; b})
    {
      if (first)
	first = NO
      else
	s = [s print ' '];
      s = [self describe s token i]
    };
  = s
}

<doc> Return the stream to which to report an error.  </doc>
OutputStream
  error
{
  num_errors++;
  global_num_errors++;
  = [err print ([lex filename], ':', [lex current_line], ": ")]
}

<doc> Return the stream to which to report an error for the {object},
    similar to {error}.  </doc>
OutputStream
  error FileElement object
{
  num_errors++;
  global_num_errors++;
  = [err print ([lex filename], ':', [object line], ": ")]
}

<doc> Return the stream to which to report a syntax error.  </doc>
OutputStream
  syntaxError
{
  OutputStream s = [[self error] print "error at "];

  = if (token >= ' ' && token <= '~')
      [s print ('\'', byte (token), '\'')]
    else
      [s print ("token ", token)]
}

<doc> Issue the error and exit.  </doc>
void
  fatalError
{
  [[self syntaxError] nl];
  [Runtime exit 1]
}

<doc> Designated initializer.  </doc>
id
  init Lexer l
{
  lex = l;
  trace = trace_default;

  [lex set_parser self];
  token = [lex token];

  = self
}

int
  nextToken
{
  if (trace)
    [err print "reading a token: "];

  token = [lex nextToken];

  if (trace)
    {
      [err print token];

      if (token >= ' ' && token <= '~')
	[err print (" '", byte (token), '\'')]
      else
	{
	  Any semantics = [lex semantics];

	  if (semantics)
	    [err print (" semantics: ", semantics)]
	};

      [err nl]
    };

  = token
}

void
  term int expected
{
  if (token != expected)
    {
      OutputStream s = [self syntaxError];

      [s print (", expected token ")];
      [self describe s token expected];
      [s nl]
    };

  [self nextToken]
}

Array
  tokens_by_number
{
  = tokens_by_number
}

end;
