/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMBind.m,v 1.5 1996/05/21 16:31:13 tiggr Exp $  */

#define OTMBIND_DECLARE_PRIVATE_METHODS
#import "OTMBind.h"
#import "OTMError.h"
#import "OTMInstance.h"
#import "OTMTuple.h"
#import "OTMTypeTuple.h"

@implementation OTMBind

+(OTMBind *) bindWithContainer: (OTMCompound *) c
{
  return [[self gcAlloc] initWithContainer: c];
}

-(void) compileCondition: (BOOL) end_not_start
{
  if (!end_not_start)
    {
      int i, n = [handlers length];

      formac (of, @"%@/* BEGIN START OF BIND  */", nl ());
      formac (of, @"%@void *%@_handler (int _i, void *condition)%@{",
	      nl (), nl (), nl ());
      indent_level++;
      formac (of, @"%@switch (_i)%@{", nl (), nl ());
      indent_level++;

      for (i = 0; i < n; i++)
	{
	  OTMExpr *handler = [[handlers _elementAtIndex: i] cdr];

	  indent_level--;
	  formac (of, @"%@case %d:", nl (), i);
	  indent_level++;

	  [handler compile];
	  formac (of, @"%@return %@;", nl (), [handler result]);

	}

      indent_level--;
      formac (of, @"%@default:", nl ());
      indent_level++;
      formac (of, @"%@abort ();%@return 0;", nl (), nl ());

      indent_level--;
      formac (of, @"%@}", nl ());
      indent_level--;
      formac (of, @"%@}", nl ());

      formac (of, @"%@struct trt_bind *_bind_info = alloca (sizeof *_bind_info"
	      @" + %d * sizeof _bind_info->classes[0]);", nl (), n);

      formac (of, @"%@_bind_info->c.kind = TRT_CONDITION_KIND_BIND;"
	      @"%@_bind_info->num = %d;%@_bind_info->handler = _handler;",
	      nl (), nl (), n, nl ());

      for (i = 0; i < n; i++)
	{
	  OTMExpr *c = [[handlers _elementAtIndex: i] car];

	  [c compile];

	  formac (of, @"%@_bind_info->classes[%d] = %@;", nl (), i, [c result]);
	}

      formac (of, @"%@trt_register (&_bind_info->c);", nl ());
      formac (of, @"%@/* FINISH START OF BIND  */\n", nl ());
    }
  else
    {
      formac (of, @"\n%@/* BEGIN END OF BIND  */", nl ());
      formac (of, @"%@trt_unwind (&_bind_info->c);", nl ());
      formac (of, @"%@/* FINISH END OF BIND  */", nl ());
    }
} /* -compileCondition: */

-(void) gcReference
{
  MARK (handlers);

  [super gcReference];
} /* -gcReference */

-(id) precompile
{
  [tom_condition_instance precompile];

  return [super precompile];
}

-(void) setHandlers: (TLCons *) hl
{
  static TLCons *ecc_tuple;

  if (!ecc_tuple)
    ecc_tuple = CONS (([OTMTypeTuple typeTupleWithSequence:
			[TLVector vectorWithElements: 2,
			 tom_condition_class_instance, the_any_ref_type]]),
		      nil);

  handlers = [TLVector vector];

  if (hl == (id) an_error)
    ;
  else if (!hl)
    warning (@"no handlers in bind");
  else while (hl)
    {
      OTMExpr *ch, *c, *h;
      TLVector *v;

      DECONS (hl, ch, hl);

      ch = resolve_expr (ch, ecc_tuple, nil, [current_either semantics]);
      if ([ch isTuple])
	{
	  v = [(OTMTuple *) ch elements];

	  c = [[v _elementAtIndex: 0] elaborate];

	  current_compound = [OTMCompound compoundWithContainer:
			      current_compound];

	  h = [[v _elementAtIndex: 1] elaborate];
	  emit_statement (h);
	  [current_compound setValue: [h value]];

	  current_compound = [current_compound container];

	  [handlers addElement: CONS (c, h)];
	}
    }
}

@end
