/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMCatch.m,v 1.6 1996/09/02 22:39:20 tiggr Exp $  */

#define OTMCATCH_DECLARE_PRIVATE_METHODS
#import "OTMCatch.h"
#import "OTMBasic.h"

@implementation OTMCatch

+(OTMCatch *) catchWithTag: (OTMExpr *) t container: (OTMCompound *) c
{
  return [[self gcAlloc] initWithTag: t container: c];
} /* +catchWithTag:container: */

-(void) compileCondition: (BOOL) end_not_start
{
  if (!end_not_start)
    {
      id <TLString> fn = ([result type] == basic_type[BT_VOID]
			  ? @"" : [[result type] frobnicatedName]);
      LTTSelArgTypes *rd = [LTTSelArgTypes argumentTypesWithString: fn];
      int n = [result flatElementCount];

      formac (of, @"%@/* BEGIN START OF CATCH  */"
	      @"%@jmp_buf _catch_buf;%@if (!setjmp (_catch_buf))%@{",
	      nl (), nl (), nl (), nl ());
      indent_level++;

      formac (of, @"%@struct trt_catch *_catch_info = alloca (sizeof "
	      @"*_catch_info + %d * sizeof _catch_info->values[0]);",
	      nl (), n);

      formac (of, @"%@_catch_info->c.kind = TRT_CONDITION_KIND_CATCH;", nl ());
      [tag compile];
      formac (of, @"%@_catch_info->tag = %@;", nl (), [tag result]);
      formac (of, @"%@_catch_info->jmp = &_catch_buf;", nl ());
#if STACK_REF_STRUCT
      formac (of, @"%@_catch_info->sp = %@;", nl (), TO_TRT_PROTECT_STACK);
#endif
      formac (of, @"%@_catch_info->value_desc = &%@;",
	      nl (), [rd outputDefinitionName]);

      if (n)
	{
	  if (n != 1)
	    ABORT (); /* XXX Closures would be nice here...  */

	  formac (of, @"%@_catch_info->values[0] = &%@;",
		  nl (), [result outputReference]);
	}

      formac (of, @"%@trt_register (&_catch_info->c);", nl ());
      formac (of, @"%@/* FINISH START OF CATCH  */\n", nl ());
    }
  else
    {
      formac (of, @"\n%@/* BEGIN END OF CATCH  */", nl ());
      formac (of, @"%@trt_unwind (&_catch_info->c);", nl ());
      indent_level--;
      formac (of, @"%@}", nl ());
      formac (of, @"%@/* FINISH END OF CATCH  */", nl ());

    }
} /* -compileCondition: */

-(void) gcReference
{
  MARK (tag);

  [super gcReference];
} /* -gcReference */

-(id) initWithTag: (OTMExpr *) t container: (OTMCompound *) c
{
  [super initWithContainer: c];

  tag = t;

  return self;
} /* -initWithTag:container: */

-(id) precompile
{
  id <TLString> fn = ([result type] == basic_type[BT_VOID]
		      ? @"" : [[result type] frobnicatedName]);
  LTTSelArgTypes *rd = [LTTSelArgTypes argumentTypesWithString: fn];

  /* XXX This is wrong, since the information should be output to the .i
     file as well.  For now, we just hope that some method somewhere in
     every program being resolved, will incorporate a method with
     arguments of the same type as this tuple we're returning...
     Fri May  3 16:19:45 1996,  --Tiggr.  */
  formac (of, @"%@extern struct trtd_selector_args %@;",
	  nl (), [rd outputDefinitionName]);

  return [super precompile];
}

@end
