/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMCustomMethod.h,v 1.11 1996/03/26 14:16:38 tiggr Exp $  */

#import "OTMMethod.h"

@interface OTMCustomMethod: OTMMethod
{
  /* Iff !nil, the definition is a foreign one.  This specifies how
     foreign.  Most of the time probably "C".  */
  id <TLString> foreign;

  /* The body of this method, if it was defined (and not foreign).  */
  OTMCompound *body;

  /* Extensions of SELF for which the variable struct pointer has been
     output.  */
  TLSet *declared_extension_pointers;

  /* Iff !0, all arguments referencing objects.  */
  TLVector *referencing_args;

  /* Iff !0, the expression of temporaries containing the return value.  */
  OTMExpr *return_value;

  /* Iff !0, the function epilogue should unprotect the stack.  */
  unsigned int stack_protected: 1;

  /* Iff !0, a return statement is present somewhere in the body of this
     method.  */
  unsigned int have_return: 1;
}

#if CACHE_EXTENSION_POINTER
-(void) compileExtensionPointerDefinitions;
-(void) declareExtensionPointer: (OTMExtension *) ext;
#endif

-(id <TLString>) foreign;
-(void) setForeign: (id <TLString>) f;

-(OTMCompound *) body;

-(void) setBody: b;

-(TLVector *) referencingArguments;

-(OTMExpr *) returnValue;
-(void) setReturnValue: (OTMExpr *) e;

-(void) setHaveReturnStatement: (BOOL) ynp;
-(BOOL) haveReturnStatement;

-(void) compile;

-(BOOL) stackProtected;
-(void) setStackProtected;

#ifdef OTMCUSTOMMETHOD_DECLARE_PRIVATE_METHODS

/* Return the result of formac'ing to S a cast, sans parenthesis, to this
   method's output function type.  */
-(id) outputCast: (id) s
   forInvocation: (OTMInvocation *) inv;

-(id) outputStartOfInvocation: (OTMInvocation *) inv;

#endif

@end
