/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMDynamicType.m,v 1.10 1996/06/10 15:37:46 tiggr Exp $  */

#import "OTMDynamicType.h"
#import "OTMInvocation.h"
#import "OTMMethod.h"

/* The single instance of the dynamic type.  */
OTMDynamicType *the_dynamic_type;

/* Dictionary from selector, involved with a dynamic type, to a dictionary
   from invocation to actual selector.  */
static TLDictionary *dynamic_selectors;

@implementation OTMDynamicType

+(void) init
{
  the_dynamic_type = [[self gcAlloc] init];
  [the_dynamic_type gcLock];
} /* +init */

+(void) addInvocation: (OTMInvocation *) inv
    toDynamicSelector: (LTTSelector *) sel
{
  TLDictionary *d;

  if (dynamic_selectors)
    d = [dynamic_selectors objectForKey: sel];
  else
    {
      dynamic_selectors = [[TLDictionary gcAlloc] init];
      [dynamic_selectors gcLock];
      d = nil;
    }

  if (!d)
    {
      d = [TLDictionary dictionary];
      [dynamic_selectors setObject: d forKey: sel];
    }

  [d setObject: [OTMMethod selectorForMethod: [inv method] invocation: inv]
     forKey: inv];
} /* +addInvocation:toDynamicSelector: */

+(void) dumpInfo: (id <TLOutputStream>) inf
{
  id <TLEnumerator> ke;
  LTTSelector *sel;

  if (!dynamic_selectors)
    return;

  formac (inf, @"(dynamic-selectors");
  ke = [dynamic_selectors enumerator];
  while ((sel = [ke nextObject]))
    {
      TLDictionary *d = [dynamic_selectors objectForKey: sel];
      id <TLEnumerator> se = [d valueEnumerator];
      LTTSelector *alias;

      formac (inf, @"\n (%#", [[sel name] internal]);

      while ((alias = [se nextObject]))
	formac (inf, @"\n  %#", [[alias name] internal]);

      formac (inf, @")");
    }

  formac (inf, @")\n");
}

+(LTTSelector *) selectorForInvocation: (OTMInvocation *) inv
		     ofDynamicSelector: (LTTSelector *) sel
{
  TLDictionary *d = [dynamic_selectors objectForKey: sel];
  LTTSelector *alias = [d objectForKey: inv];

  if (!alias)
    fatal (1, @"no actual selector for invocation %# of dynamic selector %#",
	   inv, [sel name]);

  return alias;
} /* +selectorForInvocation:ofDynamicSelector: */

-(BOOL) allowedTypeForArgumentRedeclaration: (OTMType *) t
				 inSubclass: (BOOL) subclass_p
{
  return t == self;
} /* -allowedTypeForArgumentRedeclaration:inSubclass: */

-(id <TLString>) frobnicatedName
{
  return @"x";
} /* -frobnicatedName */

-(BOOL) isFullyDefinedType
{
  return NO;
} /* -isFullyDefinedType */

-(BOOL) matches: (OTMType *) t
{
  return t == self;
} /* -matches: */

-(int) matchesConvertibly: (OTMType *) t
{
  return 0;
} /* -matchesConvertibly: */

-(id) name
{
  return @"dynamic";
} /* -name */

-(id) outputTypeName
{
  // XXX YYY ZZZ
  error (@"dynamic type used in empty context (I guess)");

  return @"dynamic";
}

-(OTMType *) typeAt: (int) index in: (int *) indices
{
  return self;
}

@end
