/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMLoop.m,v 1.7 1996/04/29 19:37:14 tiggr Exp $  */

#import "OTMLoop.h"
#import "OTMBasic.h"
#import "OTMBreak.h"
#import "OTMBuiltinMethod.h"
#import "OTMITE.h"

@implementation OTMLoop

+(OTMLoop *) loopWithContainer: (id) c
{
  return [[self gcAlloc] initWithContainer: c];
} /* +loopWithContainer: */

-(void) compile
{
  formac (of, @"%@for (;;) ", nl ());

  [super compile];
} /* -compile */

-(void) compileCondition: (BOOL) end_not_start
{
  if (condition && !end_not_start == !post_not_pre)
    [condition compile];
} /* -compileCondition: */

-(id) elaborate
{
  [super elaborate];

  if (!post_not_pre && result && result != void_expr)
    {
      /* Set the default value if the loop is never executed even once.  */
      emit_assignment (result, nil_something_with_type ([result type]));
    }    

  return self;
} /* -elaborate */

-(void) gcReference
{
  MARK (condition);

  [super gcReference];
} /* -gcReference */

-(void) setBreakValue: (OTMExpr *) e
{
  [self assignResult: e];
} /* -setBreakValue: */

-(void) setCondition: (OTMExpr *) e atEnd: (BOOL) end_not_start
{
  OTMITE *ite;

  if (e)
    {
      condition = current_compound = [OTMCompound compoundWithContainer: self];
      ite = [OTMITE iteWithCondition:
	     resolve_expr (op_invocation1 (builtin_operator_name[BO_NOT], e),
			   CONS (basic_type[BT_BOOLEAN], nil), nil,
			   [current_either semantics])];
      [ite setThen: (id) [OTMBreak jumpStatement]];
      [(OTMCompound *) condition addStatement: [ite elaborate]];
      current_compound = self;
    }

  post_not_pre = end_not_start;
} /* -setCondition: */

@end
