/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMMeta.m,v 1.61 1996/08/31 20:04:50 tiggr Exp $  */

#define OTMMETA_DECLARE_PRIVATE_METHODS
#import "OTMMeta.h"
#import "OTMClass.h"
#import "OTMInstance.h"
#import "OTMExtension.h"
#import "OTMMetaRef.h"
#import "OTMObjectVar.h"
#import "OTMType.h"
#import "global.h"

@implementation LTTMeta (Semantics)

+(id) semanticsForInstance: (LTTInstance *) inst
{
  return [(OTMInstance *) [OTMInstance gcAlloc] initWithStructure: inst];
}

+(id) semanticsForClass: (LTTClass *) cls
{
  return [(OTMClass *) [OTMClass gcAlloc] initWithStructure: cls];
}

@end

@implementation OTMMeta

+(void) init
{
  [[ltt_instance_any semantics] setDefined: YES];
}

+(void) resolveIdentifiers
{
  id <TLEnumerator> e = [LTTMeta metas];
  LTTMeta *str;

  while ((str = [e nextObject]))
    [[str semantics] resolveIdentifiers];
}

-(BOOL) allowedTypeForArgumentRedeclaration: (OTMType *) tt
				 inSubclass: (BOOL) subclass_p
{
  OTMMeta *t = (id) tt;

  return t == self;
}

-(BOOL) classp
{
  return NO;
}

-(void) compileDeclaration
{
  if (![self declarationOutputp])
    {
      id refname = [self outputTypeName];
      id previous_context = output_current_context;

      output_current_context = structure;
      [self setDeclarationOutputP: YES];

      formac (of, @"typedef struct %@ *%@;\n", refname, refname);
      [self compileObjVarTypeDeclarations];

      formac (of, @"struct %@\n{\n", refname);
      [self compileObjVarDeclarations: ++search_mark];
      formac (of, @"};\n\n");

      output_current_context = previous_context;
    }
}

-(void) compileObjVarDeclarations: (int) k
{
  id <TLEnumerator> e;
  LTTExtension *x;
  LTTMeta *m;

  if (mark == k)
    return;
  mark = k;

  e = [structure stateSupers];
  while ((m = [e nextObject]))
    [[m semantics] compileObjVarDeclarations: k];

  e = [structure extensions];
  while ((x = [e nextObject]))
    {
      OTMExtension *e = [x semantics];

      /* XXX Should be done by otmextension itself.  */
      if ([e hasVariables])
	formac (of, @"struct %@ %@;\n", [x outputExtensionStructName],
		[x outputExtensionFieldName]);
    }
}

-(void) compileObjVarTypeDeclarations
{
  id <TLEnumerator> e;
  LTTExtension *x;
  LTTMeta *m;

  e = [structure stateSupers];
  while ((m = [e nextObject]))
    [[m semantics] compileDeclaration];

  e = [structure extensions];
  while ((x = [e nextObject]))
    [[x semantics] compileObjVarTypeDeclarations];
}

-(void) compileSuperReferenceDeclaration: (OTMMeta *) the_super
{
  if (!super_refs)
    super_refs = [TLSet set];

  if (![super_refs memq: the_super])
    {
      [super_refs addElement: the_super];
      formac (of, @"extern struct _es_c_tom_State *%@;\n",
	      [structure outputSuperReference: [the_super structure]]);
    }
}

-(void) description: (id <TLMutableStream>) stream
{
  [super description: stream];
  formac (stream, @" %@", ltt_meta_name (structure));
}

-(void) dumpInfo: (id <TLOutputStream>) s
{
  id <TLEnumerator> e;
  LTTMeta *m;

  formac (s, @"\n (supers");

  e = [structure stateSupers];
  while ((m = [e nextObject]))
    formac (s, @"\n  (%# %#)", [[[m unit] name] internal],
	    [[m name] internal]);

  e = [structure behaviourSupers];
  while ((m = [e nextObject]))
    formac (s, @"\n  %@", [[m semantics] typeInfo]);

  formac (s, @")");

  if (super_refs)
    {
      OTMMeta *m;

      formac (s, @"\n (super-refs");

      e = [super_refs enumerator];
      while ((m = [e nextObject]))
	formac (s, @"\n  %@", [m typeInfo]);

      formac (s, @")");
    }
}

-equal: (OTMMeta *) o
{
  return self == o ? self : nil;
}

-(int) flatElementCount
{
  return 1;
}

-(id <TLString>) foreign
{
  return foreign;
}

-(id <TLString>) flatFrobnicatedName
{
  return @"r";
}

-(id <TLString>) frobnicatedName
{
  return @"r";
}

-(BOOL) fullyLoaded
{
  return fully_loaded;
}

-(void) gcReference
{
  MARK (structure);
  MARK (reference);
  MARK (super_refs);
  MARK (foreign);

  [super gcReference];
}

-initWithStructure: (LTTMeta *) str
{
  if (![super initWithName: [[str name] internal]])
    return nil;

  structure = str;

  return self;
}

-(OTMMeta *) instance
{
  return [[structure instance] semantics];
}

-(BOOL) isFullyDefinedType
{
  return YES;
}

-(BOOL) isObjectType
{
  return YES;
}

-(OTMMeta *) itsClass
{
  return [[structure itsClass] semantics];
}

-(BOOL) matches: (OTMType *) t
{
  return [t isObjectType];
}

-(int) matchesConvertibly: (OTMType *) t
{
  return ([t isKindOf: [OTMMeta self]]
	  && (self == (id) t
	      || [structure isProperSub: [(OTMMeta *) t structure]])) ? 0 : -1;
}

-(OTMMetaRef *) metaReference
{
  if (!reference)
    ASGN_IVAR (reference, [OTMMetaRef metaRefWithMeta: self]);
  return reference;
}

-(TLCons *) methodsNamed: (TLVector *) name_parts
		  sender: (OTMMeta *) sender
		   super: (BOOL) super_p
		confined: (OTMMeta *) confined
{
  return [(super_p ? self : [self posedSelf]) methodsNamed: name_parts
	  sender: sender super: super_p confined: confined
	  list: nil mark: ++search_mark];
}

-(TLCons *) methodsNamed: (TLVector *) name_parts
		  sender: (OTMMeta *) sender
		   super: (BOOL) super_p
		confined: (OTMMeta *) confined
		    list: (TLCons *) l
		    mark: (int) k
{
  if (mark != k)
    {
      BOOL done = !super_p || !confined;
      id <TLEnumerator> e;
      LTTExtension *ext;
      OTMMeta *sm;
      LTTMeta *m;

      mark = k;

      if (!super_p)
	{
	  e = [structure extensions];
	  while ((ext = [e nextObject]))
	    l = [[ext semantics] methodsNamed: name_parts
				 sender: sender list: l];
	}

      e = [structure stateSupers];
      while ((m = [e nextObject]))
	{
	  sm = [m semantics];

	  if (!confined || confined == sm)
	    {
	      l = [sm methodsNamed: name_parts sender: sender super: 0
		      confined: nil list: l mark: mark];
	      done = YES;
	    }
	}

      e = [structure behaviourSupers];
      while ((m = [e nextObject]))
	{
	  sm = [m semantics];

	  if (!confined || confined == sm)
	    {
	      l = [sm methodsNamed: name_parts sender: sender super: 0
		      confined: nil list: l mark: mark];
	      done = YES;
	    }
	}

      if (!done)
	error (@"%@ is not a direct super of %@",
	       type_name ((id) confined), type_name ((id) sender));
    }
  return l;
}

-(int) minimumAlignment
{
  return 8 * sizeof (void *);
}

-(id <TLString>) outputCastName
{
  return @"void *";
}

-(id <TLString>) outputFunctionTypeForType
{
  return @"reference_imp";
}

-(id <TLString>) outputTypeName
{
  return [structure outputTypeName];
}

-(OTMMeta *) posedSelf
{
  LTTMeta *poser = [structure posedSelf];

  return structure == poser ? self : [poser semantics];
}

-(id) precompile
{
  [self compileDeclaration];

  return self;
}

-(void) resolveIdentifiers
{
  id <TLEnumerator> e = [structure extensions];
  LTTExtension *ext;

  while ((ext = [e nextObject]))
    [[ext semantics] resolveIdentifiers: self];
}

-(OTMMethod *) searchSimilarMethod: (OTMMethod *) method
{
  return [[self posedSelf] searchSimilarMethod: method mark: ++search_mark];
}

-(OTMMethod *) searchSimilarMethod: (OTMMethod *) method mark: (int) k
{
  id <TLEnumerator> e;
  LTTExtension *ext;
  OTMMethod *found;
  LTTMeta *m;

  if (mark == k)
    return nil;
  mark = k;

  e = [structure extensions];
  while ((ext = [e nextObject]))
    {
      found = [[ext semantics] searchSimilarMethod: method];
      if (found)
	return found;
    }

  e = [structure stateSupers];
  while ((m = [e nextObject]))
    {
      found = [[m semantics] searchSimilarMethod: method mark: mark];
      if (found)
	return found;
    }

  e = [structure behaviourSupers];
  while ((m = [e nextObject]))
    {
      found = [[m semantics] searchSimilarMethod: method mark: mark];
      if (found)
	return found;
    }

  return nil;
}

-(void) setForeign: (id <TLString>) f
{
  ASGN_IVAR (foreign, f);
  [self setDefined: YES];
}

-(void) setFullyLoaded
{
  fully_loaded = 1;
}

-(BOOL) statep
{
  return has_state;
}

-(void) setStatep
{
  has_state = YES;
}

-(LTTMeta *) structure
{
  return structure;
}

-(OTMObjectVar *) searchEntityNamed: (id <TLString>) n
			     supers: (BOOL) search_supers
			      class: (BOOL) search_class
			  variables: (BOOL) vp
			  constants: (BOOL) cp
{
  OTMObjectVar *v;

  v = [[self posedSelf] searchEntityNamed: n supers: search_supers
			mark: ++search_mark variables: vp constants: cp];

  if (!v && search_class)
    v = [[(OTMMeta *) [[structure itsClass] semantics] posedSelf]
	 searchEntityNamed: n supers: search_supers mark: search_mark
	 variables: vp constants: cp];

  return v;
}

-(OTMObjectVar *) searchEntityNamed: (id <TLString>) n
			     supers: (BOOL) search_supers
			       mark: (int) k
			  variables: (BOOL) vp
			  constants: (BOOL) cp
{
  id <TLEnumerator> e;
  OTMObjectVar *v;
  LTTMeta *m;

  if (mark == k)
    return nil;
  mark = k;

  v = [self entityNamed: n variables: vp constants: cp];
  if (v)
    return v;

  if (search_supers)
    {
      e = [structure stateSupers];
      while ((m = [e nextObject]))
	{
	  OTMMeta *sem = [m semantics];

	  v = [sem searchEntityNamed: n supers: YES mark: k
		   variables: vp constants: cp];
	  if (v)
	    return v;
	}

      e = [structure behaviourSupers];
      while ((m = [e nextObject]))
	{
	  OTMMeta *sem = [m semantics];

	  v = [sem searchEntityNamed: n supers: YES mark: k
		   variables: NO constants: cp];
	  if (v)
	    return v;
	}
    }

  return nil;
}

-(id <TLString>) typeInfo
{
  return formac (nil, @"(%@ %# %#)",
		 [structure classp] ? @"class" : @"instance",
		 [[[structure unit] name] internal],
		 [[structure name] internal]);
}

-(BOOL) validCastTo: (OTMType *) type
{
  OTMMeta *m = (id) type;

  return (m == self
	  || ([m isKindOf: isa]
	      && ([structure isProperSub: [m structure]]
		  || [[m structure] isProperSub: structure])));
}

-(OTMObjectVar *) entityNamed: (id <TLString>) n
		    variables: (BOOL) vp
		    constants: (BOOL) cp
{
  id <TLEnumerator> e = [structure extensions];
  LTTExtension *ext;
  id v;

  while ((ext = [e nextObject]))
    {
      OTMExtension *sem = [ext semantics];

      if (vp)
	{
	  v = [sem variableNamed: n];
	  if (v)
	    return v;
	}
      if (cp)
	{
	  v = [sem constantNamed: n];
	  if (v)
	    return v;
	}
    }

  return nil;
}

/********** OTMExpr protocol **********/

-(id) type
{
  /* A class object is its own type.  */
  return self;
}

@end
