/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMTop.m,v 1.21 1996/05/09 16:58:43 tiggr Exp $  */

#import "OTMTop.h"

@implementation OTMTop

-(id) actualSelf: (id) context
{
  return self;
} /* -actualSelf */

-(int) compare: (id) o
{
  /* Compare for the purpose of sets and the like is `eq' testing.  */
  return !(o == self);
} /* -compare: */

-(BOOL) constantp
{
  return NO;
} /* constantp */

-(BOOL) declaredp
{
  return declared;
} /* -declaredp */

-(BOOL) declarationOutputp
{
  return !!Q_OUTPUT_DECL (qualifiers);
} /* -declarationOutputp */

-(BOOL) deferredp
{
  return !!Q_DEFERRED (qualifiers);
} /* -deferredp */

-(BOOL) definedp
{
  return defined;
} /* -definedp */

-(void) description: (id <TLMutableStream>) stream
{
  formac (stream, @" %@:%d", filename, line_number);
} /* -description: */

-(id) elaborate
{
  return self;
} /* -elaborate */

-(void) gcReference
{
  MARK (filename);
} /* -gcReference */

-init
{
  return [self initWithLine: current_line inFile: current_filename];
} /* -init */

-initWithLine: (int) line inFile: (id <TLString>) name
{
  filename = name;
  line_number = line;

  return self;
} /* -initWithLine:inFile: */

-(int) line
{
  return line_number;
} /* -line */

-(id <TLString>) filename
{
  return filename;
} /* -filename */

-(void) print: (id <TLMutableStream>) stream quoted: (BOOL) qp
{
  const char *s = class_get_class_name (isa);

  formac (stream, @"(%s %x", s, self);
  [self description: stream];
  [@")" print: stream quoted: NO];
} /* -print:quoted: */

-(id) precompile
{
  return self;
} /* -precompile */

-(otm_qualifiers) protection
{
  return Q_PROTECTION (qualifiers);
} /* -protection */

-(otm_qualifiers) qualifiers
{
  return qualifiers;
} /* -qualifiers */

-(otm_qualifiers) redo
{
  return Q_REDO (qualifiers);
} /* -redo */

-(void) refreshLocation
{
  filename = current_filename;
  line_number = current_line;
} /* -refreshLocation */

-(id) resolveInContext: (OTMMeta *) current
{
  return self;
}

-(void) setDeclarationOutputP: (BOOL) ynp
{
  qualifiers = Q_SET_OUTPUT_DECL (qualifiers, ynp ? OQ_OUTPUT_DECL : 0);
} /* -setDeclarationOutputP: */

-(void) setDeclared: (BOOL) ynp
{
  declared = !!ynp;
} /* -setDeclared */

-(void) setDefined: (BOOL) ynp
{
  defined = !!ynp;
} /* -setDefined */

-(void) setDeferred: (BOOL) ynp
{
  qualifiers = Q_SET_DEFERRED (qualifiers, ynp ? OQ_DEFERRED : 0);
} /* -setDeferred */

-(void) setProtection: (otm_qualifiers) prot
{
  qualifiers = Q_SET_PROTECTION (qualifiers, prot);
} /* -setProtection: */

-(void) setQualifiers: (otm_qualifiers) q
{
  qualifiers = q;
} /* -setQualifiers: */

-(void) setRedo: (otm_qualifiers) redo
{
  qualifiers = Q_SET_REDO (qualifiers, redo);
} /* -setRedo: */

-(void) setStatic: (BOOL) ynp
{
  qualifiers = Q_SET_STATIC (qualifiers, ynp ? OQ_STATIC : 0);
} /* -setStatic: */

-(BOOL) staticp
{
  return !!Q_STATIC (qualifiers);
} /* -staticp */

@end
