/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMTypeTuple.m,v 1.21 1996/06/13 16:04:21 tiggr Exp $  */

#import "OTMTypeTuple.h"
#import "OTMTuple.h"

@implementation OTMTypeTuple

+(OTMTypeTuple *) typeTupleWithSequence: (id) s
{
  return [[self gcAlloc] initWithSequence: s];
} /* +typeTupleWithSequence: */

+(OTMTypeTuple *) typeTupleWithTuple: (OTMTuple *) t
{
  TLVector *v = [t elements];
  int i, n = [v length];
  TLVector *vt = [TLVector vectorWithCapacity: n];

  for (i = 0; i < n; i++)
    [vt addElement: [[v _elementAtIndex: i] type]];

  return [self typeTupleWithSequence: vt];
} /* +typeTupleWithTuple: */

-(OTMTypeTuple *) actualSelf: (OTMMeta *) either
{
  int i, n = [elements length];
  TLVector *nel = [TLVector vectorWithCapacity: n];
  BOOL different = NO;

  for (i = 0; i < n; i++)
    {
      OTMType *o, *p = [elements _elementAtIndex: i];

      o = [p actualSelf: either];
      if (o != p)
	different = YES;
      [nel addElement: o];
    }
  return different ? [isa typeTupleWithSequence: nel] : self;
}

-(BOOL) allowedTypeForArgumentRedeclaration: (OTMType *) t
				 inSubclass: (BOOL) subclass_p
{
  id actually = [self tupleSingleElement];

  if (actually != self)
    return [actually allowedTypeForArgumentRedeclaration: t
		     inSubclass: subclass_p];

  if (![t isKindOf: isa])
    return NO;

  {
    TLVector *other = [(OTMTypeTuple *) t elements];
    int i, n = [other length];

    if (n != [elements length])
      return NO;

    for (i = 0; i < n; i++)
      if (![[elements _elementAtIndex: i]
	    allowedTypeForArgumentRedeclaration: [other _elementAtIndex: i]
	    inSubclass: subclass_p])
	return NO;
  }

  return YES;
} /* -allowedTypeForArgumentRedeclaration:inSubclass: */

-(void) description: (id <TLMutableStream>) stream
{
  [super description: stream];
  formac (stream, @" %#", elements);
} /* -description: */

-(TLVector *) elements
{
  return elements;
} /* -elements */

-(int) flatElementCount
{
  int i, t, n = [elements length];

  for (i = t = 0; i < n; i++)
    t += [[elements _elementAtIndex: i] flatElementCount];

  return t;
} /* -flatElementCount */

-(id <TLString>) flatFrobnicatedName
{
  int i, n = [elements length];
  TLMutableString *s = nil;

  for (i = 0; i < n; i++)
    s = formac (s, @"%@", [[elements _elementAtIndex: i] flatFrobnicatedName]);

  return formac (s, @"");
} /* -flatFrobnicatedName */

-(id <TLString>) frobnicatedName
{
  TLMutableString *s = formac (nil, @"(");
  int i, n = [elements length];

  for (i = 0; i < n; i++)
    formac (s, @"%@", [[elements _elementAtIndex: i] frobnicatedName]);

  return formac (s, @")");
} /* -frobnicatedName: */

-(void) gcReference
{
  MARK (elements);
  [super gcReference];
} /* -gcReference */

-initWithSequence: (id) s
{
  id <TLEnumerator> e = [s enumerator];
  OTMType *t;

  if (![super init])
    return nil;

  elements = [TLVector vector];

  while ((t = [e nextObject]) || [e notEndP])
    if (t)
      [elements addElement: t];
    else
      [elements addElement: the_any_type];

  return self;
} /* -initWithSequence: */

-(BOOL) isFullyDefinedType
{
  int i, n = [elements length];

  for (i = 0; i < n; i++)
    if (![[elements _elementAtIndex: i] isFullyDefinedType])
      return NO;

  return YES;
} /* -isFullyDefinedType */

-(BOOL) isTuple
{
  return YES;
} /* -isTuple */

-(BOOL) matches: (OTMType *) t
{
  if ([t isKindOf: isa])
    {
      TLVector *other = [(OTMTypeTuple *) t elements];
      int i, n = [other length];

      if (n != [elements length])
	return NO;

      for (i = 0; i < n; i++)
	if (![[elements _elementAtIndex: i]
	      matches: [other _elementAtIndex: i]])
	  return NO;

      return YES;
    }

  /* XXX This is not totally correct, in casu we're a singleton tuple and
     the other is the type of the element.  */
  return NO;
} /* -matches: */

-(int) matchesConvertibly: (OTMType *) t
{
  id actually = [self tupleSingleElement];

  if (actually != self)
    return [actually matchesConvertibly: t];

  if ([t isKindOf: isa])
    {
      TLVector *other = [(OTMTypeTuple *) t elements];
      int i, o, offset = 0, n = [other length];

      if (n == [elements length])
	{
	  for (i = offset = 0; i < n; i++)
	    {
	      o = [[elements _elementAtIndex: i]
		   matchesConvertibly: [other _elementAtIndex: i]];
	      if (o < 0)
		return -1;
	      offset += o;
	    }

	  return offset;
	}
    }

  return -1;
} /* -matchesConvertibly: */

-(BOOL) matchesExactly: (OTMType *) t
{
  if ([t isKindOf: isa])
    {
      TLVector *other = [(OTMTypeTuple *) t elements];
      int i, n = [other length];

      if (n != [elements length])
	return NO;

      for (i = 0; i < n; i++)
	if (![[elements _elementAtIndex: i]
	      matchesExactly: [other _elementAtIndex: i]])
	  return NO;

      return YES;
    }

  /* XXX This is not totally correct, in casu we're a singleton tuple and
     the other is the type of the element.  */
  return NO;
}
-(id <TLString>) name
{
  return (id) formac (nil, @"%@", self);
} /* -name */

-(id <TLString>) outputCastName
{
  int i, n = [elements length];
  TLMutableString *s = nil;

  for (i = 0; i < n; i++)
    s = formac (s, i ? @", %@" : @"%@",
		[[elements _elementAtIndex: i] outputCastName]);

  return formac (s, @"");
} /* -outputCastName */

-(id) precompile
{
  int i, n = [elements length];

  for (i = 0; i < n; i++)
    {
      id p = [elements _elementAtIndex: i];
      id o = [p precompile];

      if (o != p)
	[elements _replaceElementAtIndex: i by: o];
    }
  return self;
} /* -precompile: */

-(OTMType *) tupleFirstFlatElement
{
  return [[elements _elementAtIndex: 0] tupleFirstFlatElement];
} /* -tupleFirstFlatElement */

-(OTMType *) tupleSingleElement
{
  return [elements length] == 1 ? [elements _elementAtIndex: 0] : self;
} /* -tupleSingleElement */

-(OTMType *) typeAt: (int) index in: (int *) indices
{
  return (index < 0
	  ? self
	  : (indices[0] >= [elements length]
	     ? nil
	     : [[elements _elementAtIndex: indices[0]] typeAt: index - 1
						       in: indices + 1]));
}

@end
