/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMVariable.m,v 1.20 1996/05/23 13:32:21 tiggr Exp $  */

#define OTMVARIABLE_DECLARE_PRIVATE_METHODS
#import "OTMVariable.h"
#import "OTMDynamicType.h"
#import "OTMMethod.h"
#import "OTMType.h"
#import "OTMTuple.h"
#import "OTMTypeTuple.h"
#import "global.h"

int next_tmp;

@implementation OTMVariable

+(OTMVariable *) variableWithName: (OTMTop *) n type: (OTMType *) t
{
  return [[self gcAlloc] initWithName: n type: t];
} /* +variableWithName: */

+(OTMVariable *) temporaryVariableWithType: (OTMType *) t
{
  return [self variableWithName: formac (nil, @"_%d_", next_tmp++) type: t];
} /* +temporaryVariableWithType: */

-(BOOL) checkProperArgument: (OTMMethod *) m
{
  id p = [m argumentNamed: name];
  if (p)
    {
      error_for (self, @"redeclaration of `%@'", name);
      cerror_for (p, @"previous declaration was here");
    }
    
  if ([type isObjectType])
    [m setHaveStackReferences: self];

  return (type == the_dynamic_type);
}

-(void) compileAddressListWithFirst: (id <TLString>) f
			  separator: (id <TLString>) s
{
  formac (of, @"%@&%@", f, [self outputReference]);
}

-(void) compileAssignment: (OTMExpr *) rhs
{
  if ([(OTMType *) type isTuple])
    {
      /* We can not have a tuple type.  */
      ABORT ();
    }
  else
    {
      formac (of, @"%@%@ = %@;", nl (), [self outputReference], [rhs result]);
    }
} /* -compileAssignment: */

-(void) description: (id <TLMutableStream>) stream
{
  [super description: stream];
  formac (stream, @" name=%# type=%@", name, type_name ((id) type));
} /* -description: */

-(void) gcReference
{
  MARK (name);
  [super gcReference];
} /* -gcReference */

-(id) initWithName: (id) n type: (OTMType *) t
{
  if (![super initWithType: t])
    return nil;

  name = n;

  return self;
} /* -initWithName:type: */

-(OTMExpr *) lhsInvalid
{
  return nil;
} /* -lhsInvalid */

-(id <TLString>) name
{
  return name;
} /* -name */

-(id) outputDeclaration: (id) stream;
{
  return formac (stream, @"%@ %@", [(OTMType *) type outputTypeName],
		 [self outputName]);
} /* -outputDeclaration: */

-(id <TLString>) outputName
{
  return quote (name);
} /* -outputName */

-(id) outputReference
{
  return quote (name);
} /* -outputReference */

-(BOOL) residesInRefStruct
{
  return NO;
}

-(id <TLString>) result
{
  return [self outputReference];
} /* -result */

@end
