<copyright> Behaviour common to all classes and state instances.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: All.t,v 1.26 1996/09/05 15:00:08 tiggr Exp $</id>
    </copyright>

// The tom files use the double-slash comments extensively.  Unqualified
// comments indicate problems with the tom source proper, with the
// runtime, or interface to C (i.e. misfeature, bug, or something like
// that).  When qualified by `CCC' it indicates a problem with otmc (the
// tom compiler); when qualified by `GGG' it indicates a problem GNU CC
// has with the output generated.  In both cases, the comment will
// indicate what was supposed to be there originally (but was temporarily
// changed to get a working tom unit).

implementation class All end;

implementation instance
All: Conditions
{
  <doc> Paramount constants.  </doc>
  const TRUE = !0;
  const FALSE = !1;
  const YES = TRUE;
  const NO = FALSE;
}

/******************** forward / perform ********************/

<doc> Return an object of which the method indicated by the selector {sel}
    should be invoked.  This method is invoked if the receiving object
    does not directly respond to {sel}.  The default implementation
    returns {self}.  The object returned could be a delegate which is to
    act upon behalf of the receiving object for the intended call of the
    selector {sel}.  </doc>
All
  forwardDelegate selector sel
{
  = self
}

<doc> Have the receiving object behave as if the selector {sel} was
    invoked, with the, possibly unboxed, {arguments}.  </doc>
extern dynamic
  perform selector sel
	: Array arguments = nil;

<doc> Have the receiving object behave as if the selector {sel} was
    invoked, with the {arguments}.  If the selector {sel} accepts more
    than one argument, {arguments} should be a tuple.  The tuple-ization
    of the actual arguments to the selector {sel} and the elements of the
    {arguments} tuple is ignored.  </doc>
extern dynamic
  perform selector sel
     with dynamic arguments;

/******************** variable manipulation ********************/

<doc><h4>Variable manipulation</h4>

    The methods in this section provide simple variable manipulation
    functions given the name (as a string) of the variable to be retrieved
    or set.  </doc>

// These semantics are wrong; some condition could be signaled, or
// something else like that.  In any way, an (optional?) argument should
// be necessary to indicate the class where the variable originated.  The
// same remark applies to {setValue ofvariableNamed}.
// Tue May 14 23:20:13 1996, tiggr@tricky.es.ele.tue.nl
<doc> Retrieve the value of the variable with the indicated {name}.  If
    there is more than one variable with the same name and expected return
    type, the first is returned.  </doc>
extern dynamic
  valueOfVariableNamed ByteString name;

<doc> Set the value of the variable named {name} in the receiving object
    to the {value}.  It is an error if the type of the @code{value} does
    not exactly match the actual type of the variable: no conversion is
    performed.  </doc>
extern void
	 setValue dynamic value
  ofVariableNamed ByteString name;

/******************** behaviour / inheritance ********************/

<doc> Return whether the class of the receiving object is a subclass of
      the {class}.  </doc>
extern boolean
  isKindOf class (State) a_class;

<doc> Return YES iff the selector {sel} can be safely sent to the
    receiver.  The default implementation only checks whether the
    receiving object provides a direct implementation of the {sel}; any
    checking through an alternative {forwardDelegate} should be performed
    by the object itself.  </doc>
extern boolean
  respondsTo selector sel;

/******************** catch / throw ********************/

<doc> Throw execution to the catch specified for the receiving object,
    returning the {value}.  If the {value} is {void}, the default value
    for the type to be returned by the catch is returned.  </doc>
extern void
  throw dynamic value;

/******************** (un-) implementation ********************/

// These methods are `protected', but that is not really the desired
// semantics: The actual protection is supposed to be that, in the actual
// invocation, the only valid receiver is {self}.
// Sun May 19 17:22:49 1996, tiggr@tricky.es.ele.tue.nl
<doc> Moan about the selector {sel} not yet having been implemented by the
    receiving object.  </doc>
protected void
  unimplemented selector sel
       message: String message = nil
{
  // The default is nil, and it is substituted in that case by a default
  // message.  This is to remind me of the fact that it is impossible to
  // refer to the defaults from C.
  // Tue May 21 14:18:17 1996, tiggr@tom.es.ele.tue.nl
  [[SelectorCondition for self class unimplemented
		      message !message ? "not yet implemented" : message
		      selector sel] raise]
}

<doc> Contrairy to what the inheritance tells you about the selector {sel}
    being invokable for the receiving object, that object thinks
    otherwise.  </doc>
protected void
  shouldNotImplement selector sel
{
  [[SelectorCondition for self class unimplemented
		      message "should not implement" selector sel] raise]
}

<doc> Moan about the receiving object defining a method for the selector
    {sel}, but actually the implementation of the method by object thinks
    it should be implemented by a subclass.  </doc>
protected void
  subclassResponsibility selector sel
{
  [[SelectorCondition for self class unimplemented
		      message "subclass responsibility" selector sel] raise]
}

/******************** extensions ********************/

<doc> Return the {Extension} object of this object for the extension named
    {name}.  If {name == nil}, the main extension is returned.  </doc>
extern Extension
  extensionNamed String name
      inherited: boolean check_supers = NO;

/******************** miscellaneous ********************/

<doc> This is here solely to be able to print the address of objects, for
    debugging purposes.  </doc>
extern pointer
  address;

<doc> Return something describing this object.  </doc>
ByteString
  description
{
  MutableByteString s = [MutableByteString withCapacity 16];

  // This cast would not be necessary if we could declare the write method as
  //	actual (s) write OutputStream s
  // Obviously, in the context of methods, `id' is `actual (self)'.
  // Thu May  9 13:18:28 1996, tiggr@tom.es.ele.tue.nl
  s = MutableByteString ([self write s]);

  = s
}

<doc> The selector equivalent of `=='.  </doc>
boolean
  eq All other
{
  = self == other
}

<doc> Return {TRUE} when the receiving object considers itself equal to
    the other object.  The receiving object should be able to assume the
    other {object} is of the same kind, or at least shares with it a
    common superclass, as in the case of, for instance, {CharString} and
    {ByteString} (which are both subclasses of {String}).  </doc>
boolean
  equal id other
{
  = self == other
}

<doc> Return a hash value for the receiving object.  The default
    implementation returns some bit pattern deduced from {self}.  </doc>
extern int
  hash;

<doc> Return the class of the receiving object.  </doc>
deferred class (id)
  kind;

<doc> Return the receiving object.  </doc>
id
  self
{
  = self
}

<doc> All objects, even classes know how to (descriptively) write
    themselves to a stream. </doc>
deferred OutputStream
  write OutputStream s;

end;
