<copyright> Array class.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: Array.t,v 1.38 1996/06/11 16:40:42 tiggr Exp $</id>  </copyright>

<doc> {Array} is the superclass of all arrays; it is an {Indexed}
    {Collection} </doc>

implementation class
Array: State, Indexed

end;

implementation instance
Array
{
  <doc> The number of elements in the array.  </doc>
  public int length;

  <doc> A pointer to the elements of this array.  </doc>
  public pointer contents;
}

redefine extern void
  dealloc;

<doc> Return an {ArrayEnumerator} on the receiving array.  </doc>
Enumerator
  enumerator
{
  = [ArrayEnumerator with self]
}

<doc> Return a pointer to the first element in the range ({start}, {len}),
    and the number of elements in the range.  In case the latter is 0, the
    returned pointer will be NULL.  </doc>
deferred (pointer, int)
  pointerToElements (int, int) (start, len);

end;

<doc> The {ArrayEnumerator} enumerates any array, returning the elements
    boxed.  </doc>
implementation class
ArrayEnumerator: State, Enumerator

instance (id)
  with Array array
{
  = [[self alloc] init array]
}

end;

implementation instance
ArrayEnumerator
{
  <doc> The number of elements returned so far and the number of elements
	in the {array}.  </doc>
  int next, num;

  <doc> The actual array.  </doc>
  Array array;
}

protected id
  init Array a
{
  (array, num) = (a, [a length]);

  = self
}

(boolean, Any)
  next
{
  if (next == num)
    return (FALSE, nil);

  = (TRUE, array[next]);
  next++
}

<doc> All these {next} methods are not really necessary, as they are
    provided by the {Enumerator} behaviour.  However, binding them here
    directly greatly enhances speed and reduces memory requirements for
    non-object arrays, which now also do not need their own enumerator to
    obtain speed.  </doc>
redefine (boolean, byte)
  next
{
  if (next == num)
    return (FALSE, byte (0));

  = (TRUE, array[next]);
  next++
}

redefine (boolean, char)
  next
{
  if (next == num)
    return (FALSE, char (0));

  = (TRUE, array[next]);
  next++
}

redefine (boolean, int)
  next
{
  if (next == num)
    return (FALSE, 0);

  = (TRUE, array[next]);
  next++
}

redefine (boolean, long)
  next
{
  if (next == num)
    return (FALSE, 0);

  = (TRUE, array[next]);
  next++
}

redefine (boolean, float)
  next
{
  if (next == num)
    return (FALSE, 0.0);

  = (TRUE, array[next]);
  next++
}

redefine (boolean, double)
  next
{
  if (next == num)
    return (FALSE, 0.0);

  = (TRUE, array[next]);
  next++
}

end;
