<copyright> ByteString class.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: ByteString.t,v 1.34 1996/08/08 12:11:36 tiggr Exp $</id>
    </copyright>

implementation class
ByteString: ByteArray, String, FileUser
{
  <doc> The name of the default encoding.  </doc>
  public mutable static String default_encoding_name;

  <doc> A dictonary from encoding name to its decoding map.  </doc>
  static MutableDictionary decoding_maps;
}

redefine void
  load Array arguments
{
  // CCC Actually, these initializations could, of course, be collected by
  // the compiler into a load-like method, while written by the programmer
  // `on the spot'...
  // Wed Aug 7 14:52:26 1996, tiggr@cobra.es.ele.tue.nl
  default_encoding_name = "8859-1"
}

<doc> Return an array for decoding a {ByteString} which was encoding using
    the encoding known by the {encoding_name}.  If the {encoding_name} is
    {nil}, the {default_encoding_name} is used.  </doc>
CharArray
  mapForDecoding String encoding_name
{
  CharArray demap;

  if (!encoding_name)
    encoding_name = default_encoding_name;

  if (!decoding_maps)
    decoding_maps = [MutableDictionary new]
  else
    demap = decoding_maps[encoding_name];

  if (!demap)
    {
      // Should use the main bundle to locate the resource...
      // Wed Aug  7 15:06:38 1996, tiggr@cobra.es.ele.tue.nl
      String filename = [[MutableByteString new] print (encoding_name, ".icm")];
      File f = [File open filename input: YES flags: NOT_EXIST_RAISE];
      MutableCharArray zemap = [MutableCharArray withCapacity 256];
      MutableByteArray buf = [MutableByteArray withCapacity 512];
      int i;

      [zemap resize [zemap capacity]];
      [f readRange (0, 512) into buf];

      for (i = 0; i < 256; i++)
	zemap[i] = buf[2 * i + 1] + char (buf[2 * i] * 256);
      [zemap freeze];

      decoding_maps[encoding_name] = demap = zemap
    };

  = demap
}

end;

// Conversions to/from ByteStrings from/to CharString can be properly
// implemented as soon as the conversion tables are available.  These
// conversions are needed for equality, ordering comparisons, &c.
// Tue May 21 23:35:18 1996, tiggr@tricky.es.ele.tue.nl

implementation instance
ByteString

// This can be removed as soon as trt uses tom io instead of C's stdio.
// Fri May 24 16:49:43 1996, tiggr@cobra.es.ele.tue.nl
(pointer, int)
  byteStringContents
{
  = (contents, length)
}

redefine boolean
  equal String other
{
  = [other equalByteString self]
}

extern boolean
  equalByteString ByteString other;

boolean
  equalCharString CharString other
{
  [self unimplemented cmd]
}

redefine boolean
  equalUniqueString UniqueString other
{
  = [other equalByteString self]
}

<doc> Initialize the newly allocated instance with the {num} bytes at {p}.
    The receiving instance will `own' the memory at {p}.  </doc>
protected id
  init (pointer, int) (p, num)
{
  (contents, length) = (p, num);

  = self
}

extern MutableByteString
  mutableSubstring (int, int) (start, len);

extern (int, int)
  searchRegex String regex
       range: (int, int) (start, len) = (0, -1)
   fold-case: boolean fold = NO;

extern ByteString
  substring (int, int) (start, len);

extern UniqueByteString
  uniqueString;

redefine OutputStream
  write OutputStream s
{
  [s writeBytes self];

  = s
}

/******************** encodings ********************/

redefine id
  stringByDecoding String encoding_name
{
  = [self stringByDemapping [isa mapForDecoding encoding_name]]
}

id
  stringByDemapping CharArray demap
{
  MutableCharString new = [MutableCharString withCapacity length];
  int i;

  /* If the {demap} is {nil}, something went wrong and we revert to the
     identical mapping (strongly resembling ISO 8859-1).  */
  if (demap)
    for (i = 0; i < length; i++)
      new[i] = {char c = demap[self[i]]}
  else
    for (i = 0; i < length; i++)
      new[i] = {char c = self[i]};

  if (![self mutable])
    [new freeze];

  = new
}

end;
