<copyright> Dictionary and MutableDictionary classes.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: Dictionary.t,v 1.4 1996/08/31 20:03:33 tiggr Exp $</id>
    </copyright>

/******************** Dictionary ********************/

implementation class
Dictionary: HashTable, Mapped

end;

implementation instance
Dictionary
      
<doc> For the enumerator which previously was at the {element} in the
    {bucket}, return the next element and bucket, prepended by whether there
    is a next and the value object in the bucket element returned.  </doc>
extern (boolean, Any, int, pointer)
  nextValue (int, pointer) (bucket, element);

Enumerator
  keyEnumerator
{
  = [DictionaryKeyEnumerator with self]
}

Enumerator
  valueEnumerator
{
  = [DictionaryValueEnumerator with self]
}

redefine OutputStream
  write OutputStream s
{
  Enumerator e = [self keyEnumerator];
  boolean b;
  All o;

  s = [s print ('(', [isa name])];

  while ({(b, o) = [e next]; b})
    s = [s print (" (", o, ' ', self[o], ')')];

  = [s print ')']
}

// The use of these multiple-asterisks comments should be standardized.
// Sat May 25 16:15:38 1996, tiggr@tricky.es.ele.tue.nl
/********** Enumerable **********/

Enumerator
  enumerator
{
  = [self valueEnumerator]
}

id
  initWithEnumerator Enumerator e
{
  //
  [self unimplemented cmd];

  = self
}

/********** Keyed **********/

<doc> Return the value stored for the {key} if present, or {nil}
    otherwise.  </doc>
extern Any
  at All key;

/********** Mapped **********/

/********** garbage collection **********/

redefine extern void
  gc_mark_elements;

end;

/******************** MutableDictionary ********************/

implementation class
MutableDictionary: Dictionary, MutableMapped

end;

implementation instance
MutableDictionary

/********** MutableKeyed **********/

<doc> Remove the mapping for the {key}.  </doc>
extern void
  remove All key;

/********** MutableMapped **********/

extern void
  set All value
   at All key;

end;

/******************** DictionaryEnumerator ********************/

implementation class
DictionaryEnumerator: State, Enumerator

instance (id)
  with Dictionary dict
{
  = [[self alloc] init dict]
}

end;

implementation instance
DictionaryEnumerator
{
  <doc> The dictionary we're enumerating.  </doc>
  Dictionary dict;

  <doc> The bucket to visit next.  </doc>
  int bucket;

  <doc> Sneaky pointer into the {dict}'s data structure pointing to the last
      bucket element we visited.  </doc>
  pointer element;
}

id
  init Dictionary s
{
  dict = s;

  = self
}

end;

/******************** DictionaryKeyEnumerator ********************/

implementation class
DictionaryKeyEnumerator: DictionaryEnumerator

end;

implementation instance
DictionaryKeyEnumerator

/********** Enumerator **********/

(boolean, Any)
  next
{
  boolean b;
  Any object;

  (b, object, bucket, element) = [dict nextKey (bucket, element)];

  = (b, object)
}

end;

/******************** DictionaryValueEnumerator ********************/

implementation class
DictionaryValueEnumerator: DictionaryEnumerator

end;

implementation instance
DictionaryValueEnumerator

/********** Enumerator **********/

(boolean, Any)
  next
{
  boolean b;
  Any object;

  (b, object, bucket, element) = [dict nextValue (bucket, element)];

  = (b, object)
}

end;
