<copyright> File class.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: File.t,v 1.9 1996/09/07 20:02:48 tiggr Exp $</id>
    </copyright>

// Experiment #n: Constants for users of the File class (compare with the
// Trie name).
implementation class
FileUser
{
  <doc> h{Open flags}  </doc>
  <doc> If the file exists, do nothing (and return nil).  </doc>
  const EXIST_NOTHING = 1;

  <doc> If the file exists, raise a condition.  </doc>
  const EXIST_RAISE = 2;

  <doc> If the file exists and {output_p}, truncate it to zero length.  </doc>
  const EXIST_TRUNCATE = 4;

  <doc> If the file exists and {output_p}, a new version of the file will
      be created (by unlinking the old file first).  </doc>
  const EXIST_SUPERSEDE = 8;

  <doc> If the file does not exist, do nothing (and return nil).  If only
      {input_p}, {nil} will also be returned if it can't be opened anyway.
      </doc>
  const NOT_EXIST_NOTHING = 16;

  <doc> If the file does not exist, raise a condition.  </doc>
  const NOT_EXIST_RAISE = 32;

  <doc> If the file does not exist and {output_p}, create it.  </doc>
  const NOT_EXIST_CREATE = 64;

  <doc> Every write will append to the end of the file.  </doc>
  const APPEND = 128;

  <doc> A mask for the above flags.  </doc>
  const MASK = 0xff;
}

end;

implementation instance FileUser end;

implementation class
File: ByteStream, FileUser
{
  // These should be expressions, but expressions are not yet handled by
  // the compiler/resolver.
  // Tue May 21 12:46:39 1996, tiggr@tom.es.ele.tue.nl
  const OPEN_INPUT = 0x100;
  const OPEN_OUTPUT = 0x200;
}

<doc> Return the directoryname of the directory containing the file named
    {filename}.  </doc>
String
  directory-of-file String name
{
  MutableString dir = [name mutableSubstring (0, -1)];
  int i, n = [dir length];

  for (i = n - 1; i >= 0; i--)
    if (dir[i] == '/')
      break;

  = (i < 0 ? "" : i == 0 ? "/" : {[dir resize i + 1]; dir})
}

<doc> Expand the {filename} relative to the {directory}.  If {directory}
    is {nil}, expansion is relative to the current working directory.
    </doc>
String
  expand-filename String filename
     relative-to: String directory = nil
{
  // ~-Expansion and `//' recognition are not yet performed.
  // Fri Aug 23 14:31:23 1996, tiggr@cobra.es.ele.tue.nl
  if (directory != nil && [directory length] != 0 && filename[0] != '/')
    {
      MutableByteString f = [[MutableByteString new] print directory];

      if (f[[f length] - 1] != '/')
	[f print '/'];
      filename = [f print filename]
    };
  = filename
}

<doc> Express the {filename} in terms relative to the {directory}.  </doc>
String
  express-filename String filename
       relative-to String directory
{
  if (!directory || ![directory length])
    return filename;

  int fl = [filename length], dl = [directory length];
  int min = fl > dl ? dl : fl;
  int i, j, last_slash = -1;
  char c, d;

  for (i = 0; i < min; i++)
    {
      c = filename[i];
      d = directory[i];

      if (c != d)
	break;
      if (c == '/')
	last_slash = i
    };

  if (last_slash == -1)
    // Actually, this string should be of the same class as the filename.
    // Fri Sep 6 11:26:07 1996, tiggr@cobra.es.ele.tue.nl
    return [[MutableByteString new] print (directory, filename)];

  if (last_slash == dl - 1)
    return [filename substring (dl, -1)];

  MutableByteString r = [MutableByteString new];
  for (j = last_slash + 1; j < dl; j++)
    {
      c = directory[j];
      if (c == '/')
	[r print "../"]
    };

  [r print [filename substring (last_slash + 1, -1)]];

  = r
}

<doc> Return a new {File}.  </doc>
instance (id)
          open String name
         input: boolean input_p = FALSE
        output: boolean output_p = FALSE
	 flags: int action = 0
{
  = [[self alloc] init name flags ((action & MASK)
				   | (input_p ? OPEN_INPUT : 0)
				   | (output_p ? OPEN_OUTPUT : 0))]
}

<doc> Search for the file along the {path}.  The last element in the
    {path} is tried first.  </doc>
instance (id)
	 open String name
    alongPath Indexed path
   trySubdir: String subdir = nil
       flags: int action = 0
{
  int i, n = [path length];

  for (i = n - 1; i >= 0; i--)
    {
      String nm = [self expand-filename name relative-to: path[i]];
      File f = [[self alloc] init nm flags OPEN_INPUT];
      if (f != nil)
	return f
    };

  if (subdir != nil)
    = [self open [[MutableByteString new] print (subdir, '/', name)]
	    alongPath path flags: action]
}

end;

implementation instance
File
{
  <doc> The name of the file.  </doc>
  public String name;

  <doc> If it is to be reopenend, these are the flags indicating how to do
      so.  </doc>
  int flags;
}

<doc> Designated initializer.  </doc>
protected id
   init String n
  flags int f
{
  [super init];

  (name, flags) = (expand-filename (n), f);

  = [self reopen]
}

<doc> Return the name of the directory containing our file.  </doc>
String
  directoryName
{
  = directory-of-file (name)
}

<doc> Reopen the file according to the {flags}.  </doc>
extern id
  reopen;

end;

// implementation class
// MappedFile: ByteString, Descriptor
// 
// end;
// 
// implementation instance
// MappedFile
// 
// end;
