<copyright> Buffered streams classes.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: StreamBuffer.t,v 1.5 1996/06/10 15:38:08 tiggr Exp $</id>
    </copyright>

// Switching between reading and writing does not work.
// Fri May 24 14:42:16 1996, tiggr@cobra.es.ele.tue.nl

implementation class
BufferedStream: StreamStream, InputOutputStream, Conditions
{
  <doc> The default value of the default size for the buffers of our
      instances.  </doc>
  const DEFAULT_BUFFER_SIZE = 8192;

  <doc> The default size for the buffers of our instances.  If anyone sets
      this to a negative number, s/he should be ni'd.  </doc>
  int default_buffer_size;
}

// Hmmm, we do not really want to have to type `redefine' here...
// Mon May 13 23:50:56 1996, tiggr@tricky.es.ele.tue.nl
redefine void
  load MutableArray arguments
{
  default_buffer_size = DEFAULT_BUFFER_SIZE
}

end;

implementation instance
BufferedStream
{
  <doc> The buffer we use.  </doc>
  MutableByteArray buffer;

  <doc> The number of elements in the {buffer}. </doc>
  int num;

  <doc> The index of the first character not yet handled (i.e. read or
      written).  </doc>
  int next;
}

<doc> Initialize the newly allocated instance to buffer the stream {s}
      with a buffer sized the {default_buffer_size}.  </doc>
redefine id
  init Stream s
{
  = [self init s bufferSize default_buffer_size]
}

<doc> Initialize the newly allocated instance to buffer the stream {s}
      with a buffer sized {cap}.  </doc>
id
	init Stream s
  bufferSize int cap
{
  buffer = [MutableByteArray withCapacity cap];

  = [super init s]
}

<doc> Fill the buffer by reading more bytes from the {stream}.  Return the
    number of bytes read.  </doc>
protected int
  readBuffer
{
  if (next == num)
    next = num = 0;

  int n = [stream readRange (num, [buffer capacity] - num) into buffer];
  num += n;

  = n
}

<doc> Write any bytes needing to be output to the {stream}.  Return the
    number of bytes written.  </doc>
protected int
  writeBuffer
{
  int n;

  if (next != num)
    {
      n = [stream writeRange (next, num - next) from buffer];
      next += n;
      if (next == num)
	next = num = 0
    };

  = n
}

/******************** OutputStream ********************/

<doc> Flush any bytes buffered to the stream this instance is buffering.
      </doc>
redefine id
  flushOutput
{
  [self writeBuffer];
  [stream flushOutput];

  = self
}

void
  write byte b
{
  int i = [self write b];

  if (!i)
    [[Condition for self class stream-error message "write failed"] raise]
}

int
  write byte b
{
  int cap = [buffer capacity];

  if (next == cap)
    [self writeBuffer];

  if (next == cap)
    return 0;

  buffer[num++] = b;
  = 1
}

int
  writeRange (int, int) (start, length)
        from ByteArray source
{
  int n, in_buffer = num - next;

  (start, length) = [source adjustRange (start, length)];

  if (length != 0)
    {
      int cap = [buffer capacity];

      if (cap < in_buffer + length)
	in_buffer -= [self writeBuffer];

      if (length > cap)
	[self unimplemented cmd message: "writes larger than buffer size"]
      else
	num += [source writeRange (start, length) into buffer at num]
    };

  = n
}

/******************** InputStream ********************/

byte
  read
{
  int b = [self read];

  if (b == -1)
    [[Condition for self class stream-eos message "end of stream"] raise];

  = byte (b)
}

int
  read
{
  int in_buffer = num - next;

  if (in_buffer == 0)
    in_buffer = [self readBuffer];

  = in_buffer != 0 ? {byte b = buffer[next++]} : -1
}

int
  readRange (int, int) (start, length)
       into MutableByteArray destination
{
  int n, in_buffer = num - next;

  (start, length) = [destination adjustMutableRange (start, length)];

  if (!in_buffer && length <= [buffer capacity])
    in_buffer = [self readBuffer];

  if (length <= in_buffer)
    {
      n = [buffer writeRange (next, length) into destination at start];

      next += n
    }
  else
    [self unimplemented cmd message: "reads larger than buffer size"];

  = n
}

end;
