<copyright> String abstract class.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: String.t,v 1.17 1996/09/06 16:17:15 tiggr Exp $</id>  </copyright>

implementation class
String: Array, Comparable

end;

implementation instance
String

/******************** equality ********************/

<doc> Compare the receiving {String} with the other {String}.  </doc>
redeclare deferred boolean
  equal String other;

<doc> Compare the receiving {String} with the other {ByteString}.  </doc>
deferred boolean
  equalByteString ByteString other;

<doc> Compare the receiving {String} with the other {CharString}.  </doc>
deferred boolean
  equalCharString CharString other;

<doc> Compare the receiving {String} with the other {UniqueString}.  </doc>
deferred boolean
  equalUniqueString UniqueString other;

/******************** Comparable ********************/

<doc> Compare the receiving {String} with the other.  </doc>
int
  compare id other
{
  int other_len = [other length];
  int min = length > other_len ? other_len : length;
  int i;

  for (i = 0; i < min; i++)
    {
      char c = self[i];
      char d = other[i];

      if (c != d)
	return c > d ? 1 : -1
    };

  return length > other_len ? 1 : length == other_len ? 0 : -1
}

/******************** searching ********************/

<doc> Return the range of the occurrence of the {string} in the receiving
    string.  Return a negative length in case it could not be found.  The
    optional {start} and {length} can be specified to restrict the
    searching within the receiving string.  </doc>
(int, int)
  rangeOfString String string
	 range: (int, int) (start, len) = (0, -1)
{
  int i, j, l = [string length];

  (start, len) = [self adjustRange (start, len)];

  if (!l)
    return (start, 0);

  char first = string[0];
  int last = start + len - l;
  for (i = start; i <= last; i++)
    if (first == self[i])
      {
	for (j = 1; j < l; j++)
	  if (self[i + j] != {char x1 = string[j]})
	    break;
	if (j == l)
	  return (i, l)
      };

  = (start, -1)
}

<doc> Search for the regular expression {regex} in the receiving string,
    in the range indicated by {(start, len)}.  Return the range of the
    whole match, or (-1, 0), which is a non-existing zero-length substring
    outside the string, if the match failed.  </doc>
deferred (int, int)
  searchRegex String regex
       range: (int, int) (start, len) = (0, -1)
   fold-case: boolean fold = NO;

/******************** retrieving (sub-) strings ********************/

<doc> Return a (mutable) {Array} of strings, taken from the receiving
    string by splitting it at characters with the indicated {char} value.
    Thus, splitting `/usr/tmp' at each `/' returns an array holding the
    empty string, `usr', and `tmp'.

    The optional argument {limit} specifies the maximum number of items in
    which the caller is interested, or -1 for all items.  For example, if
    `/usr/foo/bar' is split on `/' in two items, the array returned
    contains `usr' and `foo/bar'.

    The optional argument {excl}, if YES specifies that zero-length
    substrings are not to be included in the result.

    The optional selector {sel} specifies the method to be called to
    extract the substrings from the receiving string.  </doc>
MutableArray
  componentsSeparatedBy char c
		 limit: int limit = -1
	  excludeEmpty: boolean excl = NO
     substringSelector: selector sel = selector ("r_substring_(ii)")
{
  MutableArray a = [MutableObjectArray new];
  int s, i;

  for (s = 0; s < length && (limit < 0 || limit > 1); s = i)
    {
      for (i = s; i < length && self[i] != c; i++)
	void;
      if (!excl || i != s)
	{
	  [a add {All x1 = [self perform sel with (s, i - s)]}];
	  limit--
	};
      if (++i == length && limit != 0 && !excl)
	{
	  [a add {All x2 = [self perform sel with (i, 0)]}];
	  limit = 0
	}
    };

  if (limit == 1)
    [a add {All x3 = [self perform sel with (s, -1)]}];

  = a
}

<doc> Return a {MutableString} holding the characters from the receiving
    {String} in the (clipped) range ({start}, {len}).  </doc>
deferred MutableString
  mutableSubstring (int, int) (start, len);

<doc> Return a {String} holding the characters from the receiving {String}
    in the (clipped) range ({start}, {len}).  </doc>
deferred String
  substring (int, int) (start, len);

<doc> Return a unique version of the receiving string.  </doc>
deferred UniqueString
  uniqueString;

/******************** case conversion ********************/

// These case conversions are rather 7-bit ASCII oriented...
// Fri Aug 16 09:29:07 1996, tiggr@cobra.es.ele.tue.nl

id
  downcase
{
  int i, n = [self length];
  MutableByteString result;

  for (i = 0; i < n; i++)
    {
      byte c = self[i];

      if (c >= 'A' && c <= 'Z')
	{
	  if (!result)
	    result = MutableByteString ([self mutableSubstring (0, -1)]);
	  result[i] = c + ' '
	}
    };

  = !result ? self : {if (![self mutable]) [result freeze]; result}
}

id
  upcase
{
  int i, n = [self length];
  MutableByteString result;

  for (i = 0; i < n; i++)
    {
      byte c = self[i];

      if (c >= 'a' && c <= 'z')
	{
	  if (!result)
	    result = MutableByteString ([self mutableSubstring (0, -1)]);
	  result[i] = c - ' '
	}
    };

  = !result ? self : {if (![self mutable]) [result freeze]; result}
}

/******************** conversions ********************/

<doc> Convert the number contained in the receiving string from index
    {start}, running for {len} bytes (which -1 for unlimited length).

    The value returned is a tuple (extracted number, occupied full range,
    actual length).  If the actual length is 0, the extracted number will
    be 0.

    The {base} defaults to 10, but can be any number.  If it is larger
    than 10, alpha characters encountered have the value of 11 + the
    offset from the alpha character to the start of its range.  Thus, `a'
    is 11, `z' is 36.

    Iff {signs}, a leading `+' or `-' sign is accepted.

    Iff {c_bases}, a number starting with a `0' denotes an octal number; a
    number starting with `0x' or `0X' is a hexadecimal number.

    Iff the {base_separator} is not 0, a number can be prefixed with a
    base indication followed by the base separator to specify the base of
    the actual number to follow.  The base is read using the a decimal
    {base}, unless {decimal_base} is {FALSE}, in which case the base is
    read in the default {base}.  Thus, `10_10', with `_' as a base
    separator, returns {base} if {decimal_base} is {FALSE}, and 10 if it
    was {TRUE}.  </doc>

// Bugs: The most negative value can not be read.  All other kinds of
// overflow are not handled either.
// GGG The value returned should be a long, but long comparisons are partly
// broken on m68k-next-nextstep3 by GNU CC 2.7.2 (should report this).
// Sat May 18 23:47:27 1996, tiggr@tricky.es.ele.tue.nl

(int, boolean, int)
    integerValue (int, int) (start, len)
    defaultBase: int base = 10
      allowSign: boolean signs = YES
    allowCBases: boolean c_bases = YES
  baseSeparator: byte base_separator = '_'
    decimalBase: boolean decimal_base = YES
{
  boolean neg;
  int value;
  int i, e;
  char c;

  (start, len) = [self adjustRange (start, len)];

  if (start == length || !len)
    return (0, TRUE, 0);

  if (base < 2)
    return (0, FALSE, 0);

  i = start;
  c = self[i];
  e = start + len;

  if (signs && (c == '-' || c == '+'))
    {
      if (c == '-')
	neg = TRUE;
      if (++i != e)
	c = self[i]
    };

  if (c_bases && c == '0')
    {
      base = 8;
      if (++i != e)
	{
	  c = self[i];
	  if (c == 'x' || c == 'X')
	    {
	      base = 16;
	      if (++i != e)
		c = self[i]
	    }
	}
    };

  while (i != e)
    {
      if (c == base_separator && base_separator != byte (0))
	{
	  if (decimal_base && base != 10)
	    return [self integerValue (start, len)
			 defaultBase: 10 allowSign: signs allowCBases: c_bases
			 baseSeparator: base_separator decimalBase: YES];

	  if (++i == e)
	    return (0, TRUE, i - start);

	  boolean b;
	  int j;
	  (value, b, j) = [self integerValue (i, e - i)
				defaultBase: int (value) allowSign: NO
				allowCBases: NO baseSeparator: byte (0)];

	  return (neg ? -value : value, b, j + i - start)
	};

      int digit;

      // These `isdigit' and `isalpha' checks are uni-flawed...
      // Sat May 18 22:41:11 1996, tiggr@tricky.es.ele.tue.nl
      if (c >= '0' && c <= '9')
	digit = c - '0'
      else if (c >= 'a' && c <= 'z')
	digit = 10 + (c - 'a')
      else if (c >= 'A' && c <= 'Z')
	digit = 10 + (c - 'A')
      else
	digit = base;

      if (digit >= base)
	break;

      value = (value * base) + digit;
      if (++i != e)
	c = self[i]
    };

  if (neg)
    value = -value;

  = (value, i == e, i - start)
}

<doc> Simple front-end for {integerValue} (with default arguments).  </doc>
int
  intValue
{
  long v;

  (v,,) = [self integerValue (0, -1)];
  = int (v)
}

<doc> Simple front-end for {integerValue}, similar to {intValue}, but not
    allowing a negative value.  For a negative value entered (due to
    {integerValue} not doing overflow checking), 0 is returned.  </doc>
int
  unsignedIntValue
{
  long v;

  (v,,) = [self integerValue (0, -1) allowSign: NO];

  if (v < 0)
    v = 0;

  = int (v)
}

/******************** encodings ********************/

<doc> Return a string by decoding it assuming it was encoded using the
    encoding named by {encoding_name}.  The default implementation simply
    returns {self}.  </doc>
id
  stringByDecoding String encoding_name
{
  = self
}

end;
