<copyright> Miscellaneous behavioural classes.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: behaviours.t,v 1.18 1996/09/06 16:17:16 tiggr Exp $</id>
    </copyright>

/******************** Comparable ********************/

implementation class
Comparable: instance (All)
end;

implementation instance
Comparable: instance (All)

<doc> Return 0 if the {other} is considered equal by the receiving object.
    1 if the receiver considers himself larger, and -1 when smaller.
    </doc>
deferred int
  compare id other;

end;

/******************** Container ********************/

<doc> A container is an object which gets to mark its elements after
    normal marking has been done.  This is very usable for unique string
    tables, DO proxies, etc; actually: all cases where an object having
    become garbage implying it should be removed from its container, and
    the container itself is not allowed to reference the object in a
    normal way (since then it would never become garbage).

    In short, a the combo of garbage collector and container implements
    weak referencing.</doc>
implementation class Container
{
  <doc> The collection of all containers.  Note that this collection
      itself is a container too.  </doc>
  static MutableEqSet all_containers;
}

// What the hell is <cod>?!  It _is_ documentation, but not supposed to go
// into the distributed head files.  Do come up with a better name.
// Thu May 30 15:03:33 1996, tiggr@tom.es.ele.tue.nl
<cod> This is what the load method could look like, but actually this
    functionality is implemented in the runtime, as we want containers
    available before the very first load method, irrespective of which it
    is.

    void
      load Array arguments
    {
      all_containers = [MutableEqSet new];

      [all_containers setIsContainer YES]
    }
</cod>

end;

implementation instance Container

<doc> The container mark method.  </doc>
deferred void
  gc_container_mark_elements;

<doc> Return {TRUE} iff the receiving object is a container.  </doc>
extern boolean
  isContainer;

<doc> Set this object to be a container, or not, depending on
    {container_p}.  </doc>
extern void
  setIsContainer boolean container_p;

end;

/******************** Enumerable ********************/

implementation class
Enumerable

instance (id)
  withEnumerable Enumerable other
{
  = [self withEnumerator [other enumerator]]
}

instance (id)
  withEnumerator Enumerator e
{
  = [[self alloc] initWithEnumerator e]
}

end;

implementation instance
Enumerable: instance (All)

deferred protected id
  initWithEnumerator Enumerator e;

<doc> Return an {Enumerator} on the receiving object.  </doc>
deferred Enumerator
  enumerator;

end;

/******************** Enumerator ********************/

implementation class Enumerator end;

implementation instance
Enumerator

<doc> Return a tuple containing the next object, preceded by a boolean
    value indicating whether the end of the enumerable has been reached;
    if the boolean is {TRUE}, the end has not yet been reached.  </doc>
deferred (boolean, Any)
  next;

<doc> Default implementations for direct value retrieving enumerators.  </doc>
(boolean, byte)
  next
{
  boolean b;
  Number o;

  (b, o) = [self next];

  = b ? (b, [o byteValue]) : (b, byte (0))
}

(boolean, char)
  next
{
  boolean b;
  Number o;

  (b, o) = [self next];

  = b ? (b, [o charValue]) : (b, char (0))
}

(boolean, int)
  next
{
  boolean b;
  Number o;

  (b, o) = [self next];

  = b ? (b, [o intValue]) : (b, 0)
}

(boolean, long)
  next
{
  boolean b;
  Number o;

  (b, o) = [self next];

  = b ? (b, [o longValue]) : (b, long (0))
}

(boolean, float)
  next
{
  boolean b;
  Number o;

  (b, o) = [self next];

  = b ? (b, [o floatValue]) : (b, 0.0)
}

(boolean, double)
  next
{
  boolean b;
  Number o;

  (b, o) = [self next];

  = b ? (b, [o doubleValue]) : (b, 0.0d)
}

end;
