<copyright> Stream classes.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: streams.t,v 1.14 1996/08/15 16:43:41 tiggr Exp $</id>  </copyright>

/******************** Stream ********************/

implementation class
Stream: Conditions

end;

implementation instance
Stream: instance (All)

deferred void
  close;

end;

/******************** InputStream ********************/

implementation class
InputStream: Stream

end;

implementation instance
InputStream

<doc> Discard any unread input.  The default implementation does nothing.
    </doc>
id
  flushInput
{
  = self
}

<doc> Read a byte from the receiving stream.  This raises a {stream-eos}
    condition upon end-of-stream.  </doc>
deferred byte
  read;

<doc> Read a byte from the receiving stream.  Return -1 for end-of-stream.
    </doc>
deferred int
  read;

<doc> Read at most {num} bytes from the stream into the {buffer}, starting
    to add bytes at its current length.  Return the number of bytes
    successfully read, which is 0 for end-of-stream.  </doc>
int
  readBytes int num
       into MutableByteArray buffer
{
  = [self readRange ([buffer length], num) into buffer]
}

<doc> Read at most {num} bytes from the stream into the {buffer}, by
    writing in it from position {start}.  Return the number of bytes
    successfully read, which is 0 for end-of-stream.  </doc>
deferred int
  readRange (int, int) (start, num)
       into MutableByteArray buffer;

<doc> Read a `\n' terminated sequence of bytes and return them (without
    the `\n' at the end).  Return {nil} upon end of file (if no characters
    have been collected).  </doc>
MutableByteString
  readLine
{
  MutableByteString t;
  int c;

  if ((c = [self read]) != -1)
    for (t = [MutableByteString withCapacity 48];
	 c != -1 && c != '\n'; c = [self read])
      [t add byte (c)];

  = t
}

end;

/******************** OutputStream ********************/

implementation class
OutputStream: Stream
{
  // Sat Jun  1 00:45:22 1996, tiggr@tricky.es.ele.tue.nl
  <doc> This is the buffer used by {print base:...}.  Obviously this fails
      in a multi threaded environment.  On the other hand, not having
      stack-local object variables (i.e. objects allocated from the stack
      and not from the heap), allocating a new object for every call to
      {print base:...} is much too expensive.  It seems we desperately
      need thread-local variables, which are static variables which have a
      distinct value in each distinct thread.  Now that's an old idea...
      </doc>
  static MutableByteArray print_buffer;
}

end;

implementation instance
OutputStream

<doc> Write out any unwritten (buffered) output.  The default
    implementation does nothing.  </doc>
id
  flushOutput
{
  = self
}

<doc> Output a new line to the receiving stream.  An interactive stream
    should override this to also flush its output if it desires line based
    buffering.  </doc>
id
  nl
{
  = [["\n" write self] flushOutput]
}

extern id
  print boolean b;

id
  print byte b
{
  [self write b];

  = self
}

extern id
  print char c;

extern id
  print int i;

extern id
  print long l;

extern id
  print float f;

extern id
  print double d;

extern id
  print pointer addr;

id
  print All object
{
  if (object == nil)
    = ["*nil*" write self]
  else
    = [object write self]
}

<doc> Send {print} to {self} for each element of the tuple {x}.  </doc>
extern id
  print dynamic x;

// This method is byte-based, not char-based...
// GGG The {value} is an int which should be a long...
// Wed May 22 13:04:04 1996, tiggr@tom.es.ele.tue.nl
<doc> Output the {value} to the receiving stream.

    The optional {base} dictates the base of the representation, which
    defaults to 10.

    If the optional {space} is not 0, it is the number of positions the
    representation must at least occupy.

    If {space} is not 0, the optional {flush} dictates how the
    representation is to be flushed.  A negative value means left, 0 for
    center, and a positive value dictates a right shift.  The absolute
    value of {flush} indicates the amount of whitespace which must be
    available at the other end.

    The optional {signed} should be 0 for unsigned, or 1 for signed.  If
    it is -1 (the default) the value is assumed unsigned, unless {base}
    has its default value, 10.

    The optional {digit_10} sets the value to use for the decimal value 10
    when using a {base} exceeding that value.  </doc>
id
    print int value
    base: int base = 10
   space: int space = 0
   flush: int flush = -1
  signed: int how_signed = -1
   range: char digit_10 = char ('a')
{
  if (!print_buffer)
    print_buffer = [MutableByteString withCapacity 40]
  else
    [print_buffer truncate 0];

  if (how_signed == -1)
    how_signed = base == 10 ? 1 : 0;

  boolean negative = how_signed == 1 && value < 0;

  do
    {
      int digit = int (value % base);
      byte ch;

      /* The % can produce a negative result since VALUE is (by
         definition) signed, even for large (seemingly unsigned)
         hexadecimal values.  */
      if (digit < 0)
	digit = -digit;

      [print_buffer add ((digit < 10)
			 ? '0' + byte (digit)
			 : byte (digit_10) + (byte (digit) - byte (10)))];

      value /= base
    } while (value != 0);

  if (negative)
    [print_buffer add '-'];

  if (space == 0)
    [print_buffer reverse (0, -1)]
  else
    {
      int n = [print_buffer length];

      if (flush < 0)
	{
	  /* Flush left.  */
	  [print_buffer reverse (0, -1)];

	  while (n < space || flush < 0)
	    {
	      [print_buffer add ' '];
	      n++;
	      flush++
	    }
	}
      else if (flush > 0)
	{
	  /* Flush right.  */
	  while (n < space || flush > 0)
	    {
	      [print_buffer add ' '];
	      n++;
	      flush--
	    };

	  [print_buffer reverse (0, -1)]
	}
      else
	{
	  /* Flush center.  */
	  int spaces = space - n;
	  int left;

	  while (left < spaces / 2)
	    {
	      [print_buffer add ' '];
	      left++
	    };

	  [print_buffer reverse (0, -1)];

	  while (left < spaces)
	    {
	      [print_buffer add ' '];
	      left++
	    }
	}
    };

  [print_buffer write self];

  = self
}

<doc> Write the byte {b}, signaling a condition upon eof.  </doc>
deferred void
  write byte b;

<doc> Write the byte {b} and return the numbre of bytes actually written.
    </doc>
deferred int
  write byte b;
  
deferred int
  writeRange (int, int) (start, length)
        from ByteArray buffer;

int
  writeBytes ByteArray buffer
{
  = [self writeRange (0, -1) from buffer]
}

end;

/******************** InputOutputStream ********************/

implementation class
InputOutputStream: InputStream, OutputStream

end;

implementation instance
InputOutputStream

end;

/******************** StreamStream ********************/

implementation class
StreamStream: Stream, State

instance (id)
  with Stream s
{
  = [[self alloc] init s]
}

end;

implementation instance
StreamStream
{
  <doc> The stream to which we output and/or from which we input.  </doc>
  public InputOutputStream stream;
}

void
  close
{
  [stream close]
}

id
  init Stream s
{
  stream = InputOutputStream (s);

  = self
}

end;

/******************** stdio ********************/

implementation class
stdio
{
  static public InputStream in;
  static public OutputStream out;
  static public InputOutputStream err;
}

<doc> Close the {descriptor}.  Raises a {stream-error} on failure.  </doc>
extern void
  close int descriptor;

end;

implementation instance stdio end;
