<copyright> Unique string classes.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: unique-strings.t,v 1.6 1996/09/06 16:17:17 tiggr Exp $</id>
    </copyright>

/******************** UniqueString ********************/

implementation class
UniqueString: String
{
  <doc> The container of all unique strings.  The container mechanism will
      ensure only those strings are kept that are still needed.  </doc>
  static MutableSet strings;
}

<doc> Initialize the {strings} container.  </doc>
redefine void
  load Array arguments
{
  strings = [MutableSet new];
  [strings setIsContainer YES]
}

<doc> Return the {UniqueString} containing the same information as the
    {String} {s}.  This is the only method which can create new unique
    strings.  </doc>
instance (id)
  with String s
{
  UniqueString u = strings[s];

  if (!u)
    {
      u = [s uniqueString];
      [strings add u]
    };

  = u
}

end;

implementation instance
UniqueString
{
  <doc> The cached hash value of this string.  </doc>
  int hash;
}

<doc> Tell the {other} to compare itself to this {UniqueString}.  </doc>
redefine boolean
  equal String other
{
  = [other equalUniqueString self]
}

<doc> Return {TRUE}.  This is simple pointer comparison.  </doc>
boolean
  equalUniqueString UniqueString other
{
  = self == other
}

/********** Comparable **********/

<doc> Override compare, to return fast upon equality.  </doc>
redefine int
  compare id other
{
  self == other ? 0 : [super compare other]
}

end;

/******************** UniqueByteString ********************/

implementation class
UniqueByteString: UniqueString, ByteString

end;

implementation instance
UniqueByteString

<doc> This definition is only here because we can not direct the
    implementation of this selector to the right class.  </doc>
// Fri May 24 13:56:26 1996, tiggr@tom.es.ele.tue.nl
redefine boolean
  equal String other
{
  = [super (UniqueString) equal other]
}

<doc> The same is true for this one.  </doc>
redefine boolean
  equalUniqueString UniqueString other
{
  = [super (UniqueString) equalUniqueString other]
}

<doc> When called for the first time, hash the string, and remember the
    value.  Every next time, return the cached value.  Obviously, this
    loses for strings with a zero hash value (which keep being hashed).
    </doc>
redefine int
  hash
{
  if (!hash)
    hash = [super (ByteString) hash];

  = hash
}

end;

/******************** UniqueCharString ********************/

implementation class
UniqueCharString: UniqueString, CharString

end;

implementation instance
UniqueCharString

<doc> Redefinition since we can not redirect to the right ({UniqueString})
    super.  </doc>
redefine boolean
  equal String other
{
  = [super (UniqueString) equal other]
}

<doc> Redefinition since we can not redirect to the right ({UniqueString})
    super.  </doc>
redefine boolean
  equalUniqueString UniqueString other
{
  = [super (UniqueString) equalUniqueString other]
}

<doc> Cached hashing, just like {UniqueByteString}.  </doc>
redefine int
  hash
{
  /* Obviously, this loses for strings with a zero hash value.  */
  if (!hash)
    hash = [super (CharString) hash];

  = hash
}

end;
