/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: extension.m,v 1.12 1996/08/27 10:05:58 tiggr Exp $  */

#import "tr.h"
#import "tom/trt.h"

id <TLString>
type_encoding_of_type (id t)
{
  id n = [t symbolp] ? [(TLSymbol *) t name] : nil;

  /* XXX This is too ugly to even consider.  */
  return (!n ? @"TRT_TE_REFERENCE"
	  : [n equal: @"void"] ? @"TRT_TE_VOID"
	  : [n equal: @"boolean"] ? @"TRT_TE_BOOLEAN"
	  : [n equal: @"byte"] ? @"TRT_TE_BYTE"
	  : [n equal: @"char"] ? @"TRT_TE_CHAR"
	  : [n equal: @"int"] ? @"TRT_TE_INT"
	  : [n equal: @"long"] ? @"TRT_TE_LONG"
	  : [n equal: @"float"] ? @"TRT_TE_FLOAT"
	  : [n equal: @"double"] ? @"TRT_TE_DOUBLE"
	  : [n equal: @"pointer"] ? @"TRT_TE_POINTER"
	  : [n equal: @"selector"] ? @"TRT_TE_SELECTOR"
	  : [n equal: @"reference"] ? @"TRT_TE_REFERENCE"
	  : (id) formac (nil, @"***unknown: %@***", n));
}

id <TLString>
c_type_for_tom_type (id t)
{
  id n = [t symbolp] ? [(TLSymbol *) t name] : nil;

  /* XXX This is too ugly to even consider.  */
  return (!n ? @"tom_object"
	  : [n equal: @"void"] ? @"void"
	  : [n equal: @"boolean"] ? @"tom_byte"
	  : [n equal: @"byte"] ? @"tom_byte"
	  : [n equal: @"char"] ? @"tom_char"
	  : [n equal: @"int"] ? @"tom_int"
	  : [n equal: @"long"] ? @"tom_long"
	  : [n equal: @"float"] ? @"tom_float"
	  : [n equal: @"double"] ? @"tom_double"
	  : [n equal: @"pointer"] ? @"void *"
	  : [n equal: @"selector"] ? @"selector"
	  : [n equal: @"reference"] ? @"tom_object"
	  : (id) formac (nil, @"***unknown: %@***", n));
}

@implementation LTIExtension (tr)

-(TLCons *) addExtensionDescriptions: (TLCons *) l
{
  return l ? [l nconc: CONS (self, nil)] : CONS (self, nil);
}

-(TLCons *) addReferenceVariables: (TLCons *) l
{
  id <TLEnumerator> e = [structure variables];
  LTTVariable *v;

  while ((v = [e nextObject]))
    {
      id type = [v type];

      if ([type isKindOf: [LTIMeta class]])
	if (l)
	  [l nconc: CONS (v, nil)];
	else
	  l = CONS (v, nil);
    }

  return l;
}

-(TLCons *) compile: (id) s
	       unit: (LTTUnit *) u
      extensionList: (TLCons *) exts
{
  if ([[structure container] unit] == u)
    {
      formac (s, @"&%@,\n", [structure outputExtensionDescriptionName]);
      exts = CONS (structure, exts);
    }
  return exts;
}

-(id) compileDeclaration: (id) s
{
  if (output_var_decl)
    s = formac (s, @"struct %@ %@;\n", [structure outputExtensionStructName],
		[structure outputExtensionFieldName]);
  return s;
}

-(id) compileDefinition: (id) s
		context: (LTTMeta *) m
		  poser: (LTTMeta *) p
		   mark: (int) k
{
  if (mark == k || ![structure hasVariables])
    return s;
  mark = k;

  s = formac (s, @"{");

  if ([structure meta] == ltt_class_state && [structure isMainExtension])
    {
      LTTInstance *pi = [(LTTClass *) p instance];
      int num_supers = [pi numStateSupers] + [pi numBehaviourSupers];
      int num_subs= [pi numStateSubs] + [pi numBehaviourSubs];

      /* isa, asi */
      formac (s, @"&%@, %@", [m metaDefinitionName], @"TGC_ASI_CLASS_P");

      /* mdt, eot */
      if (app_unit)
	formac (s, @",\n{0, 0");
      else
	formac (s, @",\n{&%@%@, &%@%@", TO_MDT_PREFIX, [pi outputTypeName],
		TO_EOT_PREFIX, [pi outputTypeName]);

      /* num_instances */
      formac (s, @",\n0");

      /* name */
      {
	id <TLString> name = [[m name] internal];
	formac (s, @",\n{%#, %d}", name, [name length]);
      }

      /* rvo */
      if (app_unit)
	formac (s, @",\n0");
      else
	formac (s, @",\n&%@%@", TO_META_RVO_PREFIX, [pi outputTypeName]);

      /* instance_size */
      if (app_unit || [[pi semantics] haveDeferredMethods])
	formac (s, @",\n0");
      else
	formac (s, @",\nsizeof (struct %@)", [pi outputTypeName]);

      /* initial_asi, class_reference */
      formac (s, @",\n%@,\n(struct trt_class **) &%@",
	      default_asi_value, [m referenceName]);

      /* extensions */
      if (app_unit)
	formac (s, @",\n0");
      else
	formac (s, @",\n&%@", [pi extensionsDescriptionName]);

      /* subs */
      if (num_subs && !app_unit)
	formac (s, @",\n&_subs%@", [pi metaDefinitionName]);
      else
	formac (s, @",\n0");

      /* supers */
      if (num_supers && !app_unit)
	formac (s, @",\n&_supers%@", [pi metaDefinitionName]);
      else
	formac (s, @",\n0");

      /* mark */
      formac (s, @",\n0");

      formac (s, @"\n}");
    }

  return formac (s, @"},\n");
}

-(id) compileExtensionDescription: (id) s
{
  if (!defined)
    {
      id <TLString> name = [structure name] ? [[structure name] internal] : nil;
      int num_vars = [structure numVariables], num_statics = 0, num_methods = 0;
      TLMutableString *buf = nil;
      id static_desc = @"0";
      id <TLEnumerator> e;
      LTIVariable *v;
      LTIMethod *m;

      defined = YES;

      /* Output information on the static variables.  */
      e = [structure staticVariables];
      if (e)
	{
	  static_desc = formac (nil, @"_static%@",
				[structure outputExtensionDescriptionName]);
	  s = formac (s, @"struct trtd_static_var %@[] = {",
		      static_desc);

	  while ((v = [e nextObject]))
	    {
	      id <TLString> vn = [[v name] internal];

	      num_statics++;
	      formac (s, @"\n{{{%#, %d}, %@}, &%@},", vn, [vn length],
		      type_encoding_of_type ([v type]), [v staticName]);
	    }
	  formac (s, @"};\n");
	}

      /* Output information on each method.  */
      e = [self methods];
      while ((m = [e nextObject]))
	{
	  if (!num_methods)
	    {
	      buf = [TLMutableString mutableString];
	      formac (buf, @"\nstatic struct trtd_methods trtd_methods_%@ = "
		      @"{%d, {", [[structure name] external], [methods length]);
	    }
	  if (app_unit && [m name])
	    [m compileDeclaration: s];
	  formac (buf, @"%@{%@, (void *) &%@, &%@}",
		  num_methods ? @",\n" : @"\n", [m name] ?: @"0",
		  [[[m meta] structure] metaDefinitionName],
		  [[[m selector] structure] outputDefinitionName]);
	  num_methods++;
	}
      if (num_methods)
	formac (s, @"%@\n}};\n", buf);

      /* Declare our extension identity.  */
      if (num_vars)
	s = formac (s, @"int %@%@;\n",
		    [structure outputExtensionIdentityName],
		    app_unit ? @"" : (id) formac (nil, @" = %d", eid));

      /* Indicate the supers inherited through us.  */
      if ([metas length])
	{
	  LTIMeta *m;

	  formac (s, @"static struct trt_metas metas_%@ = {{%d, 0}, {",
		  [[structure name] external], [metas length]);
	  e = [metas enumerator];

	  while ((m = [e nextObject]))
	    formac (s, @"\n(void *) &%@,", [[m structure] metaDefinitionName]);
	  formac (s, @"\n}};\n");
	}

      /* Describe ourselves.  */
      s = formac (s, @"struct trtd_extension %@ = {\n",
		  [structure outputExtensionDescriptionName]);

      /* meta */
      formac (s, @"(void *) &%@", [[structure meta] metaDefinitionName]);

      /* extension_object */
      formac (s, @",\n0");

      /* eid_in_a_global */
      if (num_vars)
	formac (s, @",\n&%@", [structure outputExtensionIdentityName]);
      else
	formac (s, @",\n0");

      /* name */
      if (name)
	formac (s, @",\n{%#, %d}", name, [name length]);
      else
	formac (s, @",\n{0, 0}");

      /* methods */
      if (num_methods)
	formac (s, @",\n&trtd_methods_%@", [[structure name] external]);
      else
	formac (s, @",\n0");

      /* supers */
      if ([metas length])
	formac (s, @",\n&metas_%@", [[structure name] external]);
      else
	formac (s, @",\n0");

      /* state_size */
      if (num_vars)
	formac (s, @",\nsizeof (struct %@)",
		[structure outputExtensionStructName]);
      else
	formac (s, @",\n0");

      /* state_align */
      formac (s, @",\n%d", alignment);

      /* statics */
      formac (s, @",\n%@", static_desc);

      /* num_statics */
      formac (s, @",\n%d", num_statics);

      /* num_vars */
      formac (s, @",\n%d", num_vars);

      /* vars  */
      formac (s, @",\n{");

      if (num_vars)
	{
	  id <TLEnumerator> e = [structure variables];
	  BOOL first = YES;

	  while ((v = [e nextObject]))
	    {
	      id <TLString> name = [[v name] internal];

	      if (first)
		first = NO;
	      else
		formac (s, @",");
	      formac (s, @"\n{{{%#, %d}, %@}, ", name, [name length],
		      type_encoding_of_type ([v type]));
	      formac (s, @"(char *) &((struct %@ *) 0)->%@ - (char *) 0}",
		      [structure outputExtensionStructName], [v outputName]);
	    }
	}
      formac (s, @"\n}};\n");
    }

  return s;
}

-(id) compileIdentityDefinition: (id) s
{
  id <TLEnumerator> e;
  LTIVariable *v;

  e = [structure staticVariables];
  if (e)
    while ((v = [e nextObject]))
      s = formac (s, @"%@ %@;\n", c_type_for_tom_type ([v type]),
		  [v staticName]);
  
  return s;
}

-(id) compileTypeDeclaration: (id) s
{
  id qqq = [LTIMeta class];
  id <TLEnumerator> e;
  LTIVariable *v;

  if (declared)
    return s;
  declared = 1;

  if (constants)
    {
      id <TLEnumerator> e = [constants enumerator];
      TLString *key;
      id value;

      while ((key = [e nextObject]))
	{
	  value = [constants objectForKey: key]; 
	  formac (s, @"#define %@_%@ %#\n", [[structure name] external],
		  quote (key), value);
	}
    }

  if ([structure hasVariables])
    {
      e = [structure variables];
      while ((v = [e nextObject]))
	{
	  LTIMeta *m = [v type];

	  if ([m isKindOf: qqq])
	    s = [m compileDeclaration: s];
	}
    }

  e = [structure staticVariables];
  if (e)
    while ((v = [e nextObject]))
      s = formac (s, @"extern %@ %@;\n", c_type_for_tom_type ([v type]),
		  [v staticName]);

  s = formac (s, @"extern int %@;\n", [structure outputExtensionIdentityName]);

  if (output_var_decl)
    formac (s, @"struct %@ %@;\n", [structure outputExtensionStructName],
	    output_var_decl);

  return s;
}

-(BOOL) hasPointers
{
  id qqq = [LTIMeta class];
  id <TLEnumerator> e;
  LTIVariable *v;
  static id Qpointer;

  if (!Qpointer)
    Qpointer = [TLSymbol symbolWithName: @"pointer"];
  
  if ([structure hasVariables])
    {
      e = [structure variables];
      while ((v = [e nextObject]))
	{
	  id m = [v type];

	  if (m == Qpointer)
	    return YES;
	}
    }

  e = [structure staticVariables];
  if (e)
    while ((v = [e nextObject]))
      {
	id m = [v type];

	if (![m isKindOf: qqq] && ![m compare: @"pointer"])
	  return YES;
      }

  return NO;
}

@end
