/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: arraydef.c,v 1.32 1996/08/20 15:59:35 tiggr Exp $  */

tom_object
C4 (c_tom_, CP, ARRAY, _r_with_x)
  (tom_object self, selector cmd, ...)
{
  tom_object r;
  va_list ap;

  va_start (ap, cmd);
  r = array_with_elts (self, cmd, ap, ETD, sizeof (ET));
  va_end (ap);

  return r;
}

ET
C6 (i_tom_, CP, ARRAY, _, F, _at_i)
  (tom_object self, selector cmd, tom_int index)
{
  struct _es_i_tom_Array *this = trt_ext_address (self, _ei_i_tom_Array);

  if ((unsigned) index >= (unsigned) this->length)
    fatal ("bad index %d for array bound %d", (int) index, (int) this->length);

  return CONTENTS (this)[index];
}

tom_object
C4 (i_tom_Mutable, CP, ARRAY, _r_initWithCapacity_i)
  (tom_object self, selector cmd, tom_int cap)
{
  struct _es_i_tom_Array *this;
  struct _es_i_tom_MutableArray *that;

  this = trt_ext_address (self, _ei_i_tom_Array);
  that = trt_ext_address (self, _ei_i_tom_MutableArray);

  if (cap)
    CONTENTS (this) = xmalloc (cap * sizeof (ET));
  that->capacity = cap;

  return self;
}

ET
C6 (i_tom_Mutable, CP, ARRAY, _, F, _removeAt_i)
  (tom_object self, selector cmd, tom_int index)
{
  struct _es_i_tom_Array *this = trt_ext_address (self, _ei_i_tom_Array);
  ET previous;

  if ((unsigned) index >= (unsigned) this->length)
    fatal ("bad index %d for array bound %d", (int) index, (int) this->length);

  previous = CONTENTS (this)[index];
  if (this->length - index - 1)
    memmove (CONTENTS (this) + index, CONTENTS (this) + index + 1,
	     this->length - index - 1);
  this->length--;

  return previous;
}

void
C4 (i_tom_Mutable, CP, ARRAY, _v_removeElements__ii_)
  (tom_object self, selector cmd, tom_int start, tom_int length)
{
  array_remove_elements (self, start, length, sizeof (ET));
}

void
C4 (i_tom_Mutable, CP, ARRAY, _v_resize_i)
  (tom_object self, selector cmd, tom_int to)
{
  struct _es_i_tom_Array *this;
  struct _es_i_tom_MutableArray *that;

  this = trt_ext_address (self, _ei_i_tom_Array);
  that = trt_ext_address (self, _ei_i_tom_MutableArray);

  if (to < 0)
    RAISE (self, c_tom_Conditions_program_condition,
	   byte_string_with_c_string ("negative re-size"));

  if (to < this->length)
    this->length = to;
  else if (to > this->length)
    {
      if (to > that->capacity)
	{
	  this->contents = xrealloc (this->contents, to * sizeof (ET));
	  that->capacity = to;
	}
      BZERO (CONTENTS (this) + this->length, (to - this->length) * sizeof (ET));
      this->length = to;
    }
}

void
C4 (i_tom_Mutable, CP, ARRAY, _v_swap__ii_)
  (tom_object self, selector cmd, tom_int i, tom_int j)
{
  struct _es_i_tom_Array *this;

  this = trt_ext_address (self, _ei_i_tom_Array);

  if ((unsigned) i > (unsigned) this->length)
    fatal ("bad index 1 %d for array bound %d", (int) j, (int) this->length);
  if ((unsigned) j > (unsigned) this->length)
    fatal ("bad index 1 %d for array bound %d", (int) j, (int) this->length);

  if (i != j)
    {
      ET a_i = CONTENTS (this)[i];
      ET a_j = CONTENTS (this)[j];

      CONTENTS (this)[i] = a_j;
      CONTENTS (this)[j] = a_i;
    }
} /* swap () */

void
C6 (i_tom_Mutable, CP, ARRAY, _v_set_, F, _at_i)
  (tom_object self, selector cmd, ET elt, tom_int index)
{
  struct _es_i_tom_MutableArray *that;
  struct _es_i_tom_Array *this;

  this = trt_ext_address (self, _ei_i_tom_Array);
  that = trt_ext_address (self, _ei_i_tom_MutableArray);

  if ((unsigned) index > (unsigned) this->length)
    fatal ("bad index %d for array bound %d", (int) index, (int) this->length);

  if (index >= that->capacity)
    {
      tom_int new_cap = that->capacity ? 2 * that->capacity : 4;
      VOID_CONTENTS (this) = xrealloc (CONTENTS (this), new_cap * sizeof (ET));
      BZERO (CONTENTS (this) + that->capacity,
	     (new_cap - that->capacity) * sizeof (ET));
      that->capacity = new_cap;
    }

  CONTENTS (this)[index] = elt;
  if (index == this->length)
    this->length++;
}

void *
C4 (i_tom_, CP, ARRAY, __pi__pointerToElements__ii_)
     (tom_object self, selector cmd, tom_int start, tom_int len,
      tom_int *actual)
{
  struct _es_i_tom_Array *this = trt_ext_address (self, _ei_i_tom_Array);

  start = TRT_SEND ((int_imp), self, SEL (_ii__adjustRange__ii_),
		    start, len, &len);
  *actual = len;

  return len ? CONTENTS (this) + start : NULL;
}

#ifndef NO_INIT_WITH_ENUMERATOR

tom_object
C4 (i_tom_, CP, ARRAY, _r_initWithEnumerator_r)
  (tom_object self, selector cmd, tom_object e)
{
  struct _es_i_tom_Array *this = trt_ext_address (self, _ei_i_tom_Array);
  int cap = 0;
  tom_byte b;
  ET o;

  for (;;)
    {
      b = TRT_SEND ((int_imp), e, SEL (C3 (_o, F, __next)), &o);
      if (!b)
	break;

      if (this->length == cap)
	{
	  int new_cap = cap ? cap * 2 : 8;

	  VOID_CONTENTS (this) = xrealloc (CONTENTS (this),
					   new_cap * sizeof (ET));
	  BZERO (CONTENTS (this) + cap, (new_cap - cap) * sizeof (ET));
	  cap = new_cap;
	}
      CONTENTS (this)[this->length++] = o;
    }

  return self;
}

#endif

#ifdef STRING_METHODS

tom_byte
C5 (i_tom_, CP, String_o_equal, CP, String_r)
  (tom_object self, selector cmd, tom_object o)
{
  if (self == 0)
    return 1;
  else
    {
      struct _es_i_tom_Array *this = trt_ext_address (self, _ei_i_tom_Array);
      struct _es_i_tom_Array *that = trt_ext_address (o, _ei_i_tom_Array);

      return (this->length == that->length
	      && !memcmp (this->contents, that->contents,
			  this->length * sizeof (ET)));
    }
}

tom_object
C3 (i_tom_, CP, String_r_mutableSubstring__ii_)
  (tom_object self, selector cmd, tom_int start, tom_int len)
{
  tom_object str = TRT_SEND ((reference_imp),
			     C3 (_mr_c_tom_Mutable, CP, String), SEL (r_alloc));
  struct _es_i_tom_Array *this;

  start = TRT_SEND ((int_imp), self, SEL (_ii__adjustRange__ii_),
		    start, len, &len);
  this = trt_ext_address (self, _ei_i_tom_Array);

  return TRT_SEND ((reference_imp), str, SEL (r_init__pi_),
		   memcpy (xmalloc (len * sizeof (ET)),
			   (ET *) this->contents + start, len), len);
}

tom_object
C3 (i_tom_, CP, String_r_substring__ii_)
  (tom_object self, selector cmd, tom_int start, tom_int len)
{
  tom_object str = TRT_SEND ((reference_imp), C3 (_mr_c_tom_, CP, String),
			     SEL (r_alloc));
  struct _es_i_tom_Array *this;

  start = TRT_SEND ((int_imp), self, SEL (_ii__adjustRange__ii_),
		    start, len, &len);
  this = trt_ext_address (self, _ei_i_tom_Array);

  return TRT_SEND ((reference_imp), str, SEL (r_init__pi_),
		   memcpy (xmalloc (len * sizeof (ET)),
			   (ET *) this->contents + start, len), len);
}

tom_object
C3 (i_tom_, CP, String_r_uniqueString)
  (tom_object self, selector cmd)
{
  struct _es_i_tom_Array *this = trt_ext_address (self, _ei_i_tom_Array);
  tom_object str = TRT_SEND ((reference_imp), C3 (_mr_c_tom_Unique, CP, String),
			     SEL (r_alloc));

  return TRT_SEND ((reference_imp), str, SEL (r_init__pi_),
		   memcpy (xmalloc (this->length * sizeof (ET)),
			   (ET *) this->contents, this->length),
		   this->length);
}

#endif
