/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: hash.c,v 1.1 1996/07/01 16:51:47 tiggr Exp $  */

#include <string.h>
#include "trt.h"
#include <tom/tom-r.h>

int
hash_hash_pointer (void *elt)
{
  return HASH_POINTER (elt);
}

int
hash_hash_string (void *elt)
{
  char *s = elt;
  int h = 0, i;

  for (i = 0; i < 32 && s[i]; i++)
    h = (h << 3) ^ s[i];

  return h;
}

int
hash_hash_selector (selector sel)
{
  return sel->sel_id;
}

int
hash_equal_pointer (void *a, void *b)
{
  return a == b;
}

int
hash_equal_string (void *a, void *b)
{
  return a == b || !strcmp (a, b);
}

void
hash_resize (struct hashtable *ht)
{
  int h, i, n = ht->num_buckets ? ht->num_buckets * 4 : 4;
  struct bucket *c, *nc;
  struct bucket **new = xcalloc (n, sizeof (*new));

  for (i = 0; i < ht->num_buckets; i++)
    for (c = ht->buckets[i]; c; c = nc)
      {
	h = ht->hash (c->key) & (n - 1);
	nc = c->cdr;
	c->cdr = new[h];
	new[h] = c;
      }
  xfree (ht->buckets);

  ht->num_buckets = n;
  ht->buckets = new;
}

struct bucket *
hash_add (struct hashtable *ht, void *key, void *value)
{
  int hash = ht->hash (key);
  struct bucket *b;

  if (ht->num_buckets == ht->num_elts)
    hash_resize (ht);

  hash = hash & (ht->num_buckets - 1);
  for (b = ht->buckets[hash]; b; b = b->cdr)
    if (ht->equal (b->key, key))
      {
	b->value = value;
	return b;
      }

  b = xmalloc (sizeof (*b));
  b->key = key;
  b->value = value;
  b->cdr = ht->buckets[hash];
  ht->buckets[hash] = b;
  ht->num_elts++;

  return b;
}

void *
hash_find (struct hashtable *ht, void *key)
{
  struct bucket *b;
  int hash;

  if (!ht->num_buckets)
    return NULL;

  hash = ht->hash (key) & (ht->num_buckets - 1);
  for (b = ht->buckets[hash]; b; b = b->cdr)
    if (ht->equal (b->key, key))
      return b->value;

  return NULL;
}
