/* Interface to definitions for key collections (dictionary, bag, set, &c).
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: keyed.h,v 1.3 1996/05/30 14:41:36 tiggr Exp $  */

typedef struct bucket_elt
{
  /* Pointer to the next bucket_elt in the list.  */
  void *next;

  tom_object key;
} bucket_elt;

typedef struct bag_elt
{
  bucket_elt c;

  tom_int count;
} bag_elt;

typedef struct dict_elt
{
  bucket_elt c;

  tom_object value;
} dict_elt;

typedef struct set_elt
{
  bucket_elt c;
} set_elt;

/* Return the bucket for the KEY in the HashTable SELF.  Return NULL if it
   can not be found.  */
bucket_elt *bucket_elt_find (tom_object self, tom_object key);

/* Return the bucket for the KEY in the HashTable SELF.  A new bucket will
   be created of size SIZE, if it can not be found.  When created, the
   bucket returned will have its key set to the KEY.  */
bucket_elt *bucket_elt_find_create (tom_object self, tom_object key, int size);

/* Return the bucket elt for the KEY and remove it from the hashtable
   SELF.  The caller is responsible for freeing the bucket.  */
bucket_elt *bucket_elt_remove (tom_object self, tom_object key);

bucket_elt *bucket_elt_find_eq (tom_object self, tom_object key);
bucket_elt *bucket_elt_find_create_eq (tom_object self, tom_object key,
				       int size);
bucket_elt *bucket_elt_remove_eq (tom_object self, tom_object key);

/* Invoked by the garbage collector to have the containers do their work.  */
void tgc_mark_containers (void);
