/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: main.c,v 1.55 1996/08/20 15:59:36 tiggr Exp $  */

#include "trt.h"
#include <tom/tom-r.h>
#include <stdio.h>

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

/* For use from the debugger.  */
int trt_bucket_size = TRT_BUCKET_SIZE;

/* Common declarations for the runtime structures.  These can be defined
   by static resolving.  */
struct trt_method_dispatch_table *dtable_default;
int trt_num_eids, trt_search_mark;
struct trtd_selectors *trt_selectors;
struct trt_metas *trt_metas;
struct trt_dynamic_selectors *trt_dynamic_selectors;

/* The dictionary of classes.  */
tom_object trt_all_classes;

void
d (tom_object o)
{
  TRT_SEND ((reference_imp), c_tom_stdio_err, SEL (r_print_r), o);
  TRT_SEND ((reference_imp), c_tom_stdio_err, SEL (r_nl));
}

static void
idc (struct trt_class *m)
{
  fprintf (stderr, "%c %s", TGC_CLASS_P (m->asi) ? 'c' : 'm', m->info.name.s);
}

static void
dc (struct trt_class *m)
{
  idc (m);
  fprintf (stderr, "\n");
}

static void
dcl (struct trt_metas *ml)
{
  if (ml)
    {
      int i;

      for (i = 0; i < ml->st.num; i++)
	{
	  if (i)
	    fprintf (stderr, ", ");
	  idc (ml->metas[i]);
	}
      fprintf (stderr, "\n");
    }
}

static void
ide (struct trtd_extension *x)
{
  idc (x->meta);
  fprintf (stderr, " %s", x->name.s ?: "MAIN");
}

static void
de (struct trtd_extension *x)
{
  ide (x);
  fprintf (stderr, "\n");
}

static void
del (struct trt_meta_extensions *el)
{
  if (el)
    {
      int i;

      for (i = 0; i < el->st.num; i++)
	{
	  if (i)
	    fprintf (stderr, ", ");
	  ide (el->extensions[i]);
	}
      fprintf (stderr, "\n");
    }
}

void *
i_tom_All_p_address (tom_object self, selector cmd)
{
  return self;
}

tom_int
i_tom_All_i_hash (tom_object self, selector cmd)
{
  /* Shift it to remove the lower bits so it can be used in power-of-two
     sized hashtables.  */
  return ((tom_int) self) >> 3;
}

selector
trt_selector_named (char *s, int len)
{
  int i;

  /* XXX This is slow.  */
  for (i = 0; i < trt_selectors->st.num; i++)
    if (len == trt_selectors->selectors[i]->name.len
	&& !memcmp (s, trt_selectors->selectors[i]->name.s, len))
      return trt_selectors->selectors[i];

  return 0;
}

int
trt_all_arguments_size (selector sel)
{
  int size = 2 * sizeof (void *);
  int i;

  if (sel->out->num > 1)
    size += (sel->out->num - 1) * sizeof (void *);

  for (i = 0; i < sel->in->num; i++)
    switch (sel->in->args[i])
      {
      case TRT_TE_BOOLEAN:
      case TRT_TE_BYTE:
      case TRT_TE_CHAR:
      case TRT_TE_INT:
	size += sizeof (tom_int);
	break;

      case TRT_TE_LONG:
	size += sizeof (tom_long);
	break;

      case TRT_TE_FLOAT:
	size += sizeof (tom_float);
	break;

      case TRT_TE_DOUBLE:
	size += sizeof (tom_double);
	break;

      case TRT_TE_POINTER:
      case TRT_TE_SELECTOR:
      case TRT_TE_REFERENCE:
	size += sizeof (void *);
	break;

      default:
	ABORT ();
      }

  return (size + 7) & ~7;
}

void
c_tom_Runtime_v_fastExit_i (tom_object self, selector cmd, tom_int rc)
{
  exit (rc);
}

tom_object
c_tom_Runtime_r_nameOfSelector_s (tom_object self, selector cmd,
				  selector sel)
{
  return byte_string_with_string (sel->name.s, sel->name.len);
}

selector
c_tom_Runtime_s_selectorNamed_r (tom_object self, selector cmd,
				 tom_object name)
{
  selector sel;
  tom_int len;
  void *s;

  s = TRT_SEND ((pointer_imp), name, SEL (_pi__byteStringContents), &len);
  sel = trt_selector_named (s, len);

  if (!sel)
    fatal ("unknown selector `%s' (length=%d)", s, (int) len);

  return sel;
}

tom_object
c_tom_Runtime_r_classes (tom_object self, selector cmd)
{
  if (!trt_all_classes)
    {
      int i;

      trt_all_classes = TRT_SEND ((reference_imp), _mr_c_tom_MutableDictionary,
				  SEL (r_new));

      for (i = 0; i < trt_metas->st.num; i++)
	{
	  tom_object name = TRT_SEND ((reference_imp), trt_metas->metas[i],
				      SEL (r_name));
	  TRT_SEND ((reference_imp), trt_all_classes, SEL (v_set_r_at_r),
		    trt_metas->metas[i], name);
	}
    }

  return trt_all_classes;
}

tom_object
c_tom_Runtime_r_hostname (tom_object self, selector cmd)
{
  if (!c_tom_Runtime_hostname)
    {
      char *s = alloca (1024);

      /* This can only fail if S is an invalid address, which it isn't.  */
      gethostname (s, 1024);

      c_tom_Runtime_hostname = byte_string_with_c_string (s);
    }
  return c_tom_Runtime_hostname;
}

void *
trt_address_of_ivar (tom_object self, tom_object name,
		     enum trt_type_encoding type)
{
  struct trt_meta_extensions *exts = self->isa->info.extensions;
  struct trtd_extension *ext;
  void *ns;
  tom_int nl;
  int i, j;

  ns = TRT_SEND ((pointer_imp), name, SEL (_pi__byteStringContents), &nl);

  for (i = 0; i < exts->st.num; i++)
    {
      for (j = 0, ext = exts->extensions[i]; j < ext->num_vars; j++)
	if (nl == ext->vars[j].c.name.len
	    && !memcmp (ns, ext->vars[j].c.name.s, nl)
	    && type == ext->vars[j].c.type)
	  return ((char *) trt_ext_address (self, *ext->eid_in_a_global)
		  + ext->vars[j].offset);

      for (j = 0; j < ext->num_statics; j++)
	if (nl == ext->statics[j].c.name.len
	    && !memcmp (ns, ext->statics[j].c.name.s, nl)
	    && type == ext->statics[j].c.type)
	  return ext->statics[j].address;
    }

  fatal ("object 0x%x, of kind %s, has no variable named `%s' typed %s",
	 self, self->isa->info.name.s, c_string_with_length (ns, nl),
	 trt_type_name (type));
  return 0;
}

void
i_tom_All_x_valueOfVariableNamed_r (tom_object self, selector cmd,
				    tom_object name)
{
  builtin_return_type rt;
  void *address;

#ifdef DEBUG
  BZERO (&rt, sizeof (rt));
#endif

  if (cmd->out->num != 1)
    fatal ("return type of `%s' bad type for variable", cmd->name.s);

  address = trt_address_of_ivar (self, name, cmd->out->args[0]);

  switch (cmd->out->args[0])
    {
    case TRT_TE_CHAR: rt.i.i = *(tom_char *) address; break;
    case TRT_TE_INT: rt.i.i = *(tom_int *) address; break;
    case TRT_TE_LONG: rt.l.l = *(tom_long *) address; break;
    case TRT_TE_FLOAT: rt.f.f = *(tom_float *) address; break;
    case TRT_TE_DOUBLE: rt.d.d = *(tom_double *) address; break;

    case TRT_TE_BOOLEAN:
    case TRT_TE_BYTE:
      rt.i.i = *(tom_byte *) address;
      break;

    case TRT_TE_SELECTOR:
    case TRT_TE_POINTER:
    case TRT_TE_REFERENCE: rt.p.p = *(void **) address; break;

    default:
      fatal ("valueOfVariableNamed: unhandled type %s",
	     trt_type_name (cmd->out->args[0]));
    }

  __builtin_return (&rt);
}

void
i_tom_All_v_setValue_x_ofVariableNamed_r
  (tom_object self, selector cmd, ...)
{
  void *address;
  va_list ap;
  tom_object name;

  /* This is not very `clean', but it is better than having to come up
     with a strange name for this method just to have the value as a last
     argument.  */
  /* The assignments to zero are only here to make the compiler shut up.  */
  tom_byte b = 0;
  tom_char c = 0;
  tom_int i = 0;
  tom_long l = 0;
  tom_float f = 0;
  tom_double d = 0;
  void *p;
  enum trt_type_encoding te;
  
  if (cmd->in->num != 2)
    fatal ("bad argument 1 of `%s' for set variable", cmd->name.s);

  va_start (ap, cmd);;

  te = cmd->in->args[0];
  switch (te)
    {
    case TRT_TE_CHAR: c = va_arg (ap, tom_char); break;
    case TRT_TE_INT: i = va_arg (ap, tom_int); break;
    case TRT_TE_LONG: l = va_arg (ap, tom_long); break;
    case TRT_TE_FLOAT: f = va_arg (ap, tom_float); break;
    case TRT_TE_DOUBLE: d = va_arg (ap, tom_double); break;

    case TRT_TE_BOOLEAN:
    case TRT_TE_BYTE:
      b = va_arg (ap, tom_byte);
      break;

    case TRT_TE_SELECTOR:
    case TRT_TE_POINTER:
    case TRT_TE_REFERENCE:
      p = va_arg (ap, void *);
      break;

    default:
      fatal ("setValue: ofVariableNamed: unhandled type %s",
	     trt_type_name (cmd->out->args[0]));
    }

  name = va_arg (ap, tom_object);
  va_end (ap);

  address = trt_address_of_ivar (self, name, cmd->in->args[0]);

  switch (te)
    {
    case TRT_TE_CHAR: *(tom_char *) address = c; break;
    case TRT_TE_INT: *(tom_int *) address = i; break;
    case TRT_TE_LONG: *(tom_long *) address = l; break;
    case TRT_TE_FLOAT: *(tom_float *) address = f; break;
    case TRT_TE_DOUBLE: *(tom_double *) address = d; break;

    case TRT_TE_BOOLEAN:
    case TRT_TE_BYTE:
      *(tom_byte *) address = b;
      break;

    case TRT_TE_SELECTOR:
    case TRT_TE_POINTER:
    case TRT_TE_REFERENCE:
      *(void **) address = va_arg (ap, void *);
      break;

    default:
      fatal ("setValueOfVariableNamed to: : unhandled type %s",
	     trt_type_name (cmd->out->args[1]));
    }
}

tom_byte
i_tom_All_o_respondsTo_s (tom_object self, selector cmd, selector sel)
{
  return (trt_lookup (self, sel)
	  != (void (*) (tom_object, selector)) trt_forward);
}

tom_byte
trt_submeta_star (struct trt_class *super, struct trt_class *sub, int k)
{
  if (super == sub)
    return 1;
  if (super->info.mark == k)
    return 0;
  super->info.mark = k;

  if (super->info.subs)
    {
      int i, n = super->info.subs->st.num;

      for (i = 0; i < n; i++)
	if (trt_submeta_star (super->info.subs->metas[i], sub, k))
	  return 1;
    }

  return 0;
}

tom_byte
trt_supermeta_star (struct trt_class *sub, struct trt_class *super, int k)
{
  if (sub == super)
    return 1;
  if (sub->info.mark == k)
    return 0;
  sub->info.mark = k;

  if (sub->info.supers)
    {
      int i, n = sub->info.supers->st.num;

      for (i = 0; i < n; i++)
	if (trt_supermeta_star (sub->info.supers->metas[i], super, k))
	  return 1;
    }

  return 0;
}

tom_byte
i_tom_All_o_isKindOf_r (tom_object self, selector cmd, tom_object class)
{
  return trt_supermeta_star ((void *) self->isa, (void *) class,
			     ++trt_search_mark);
}

/* Search for the extension with the indicated name within this meta.  If
   !NAME, searching is to be done for the main extension.  */
static struct trtd_extension *
trt_meta_extension_named (struct trt_class *class, char *name, int len)
{
  struct trt_meta_extensions *x = class->info.extensions;

  /* _builtin_.Top does not have any extensions.  */
  if (x)
    {
      int i, n = x->st.num;

      if (!name)
	{
	  for (i = 0; i < n; i++)
	    if (!x->extensions[i]->name.len)
	      return x->extensions[i];
	}
      else
	for (i = 0; i < n; i++)
	  if (x->extensions[i]->name.len == len
	      && !memcmp (x->extensions[i]->name.s, name, len))
	    return x->extensions[i];
    }

  return 0;
}

struct trtd_extension *
trt_metas_extension_named (struct trt_class *class, char *name, int len, int k)
{
  struct trtd_extension *x;

  if (class->info.mark == k)
    return 0;
  class->info.mark = k;

  x = trt_meta_extension_named (class, name, len);
  if (x)
    return x;

  if (class->info.supers)
    {
      struct trt_metas *supers = class->info.supers;
      int i, n = class->info.supers->st.num;

      for (i = 0; i < n; i++)
	{
	  x = trt_metas_extension_named (supers->metas[i], name, len, k);
	  if (x)
	    return x;
	}
    }

  return 0;
}

tom_object
i_tom_All_r_extensionNamed_r_inherited__o (tom_object self, selector cmd,
					   tom_object name, tom_byte supers)
{
  struct trtd_extension *x;
  tom_int l = 0;
  void *s;

  s = (name ? TRT_SEND ((pointer_imp), name, SEL (_pi__byteStringContents), &l)
       : 0);

  if (!supers)
    x = trt_meta_extension_named (self->isa, s, l);
  else
    x = trt_metas_extension_named (self->isa, s, l, ++trt_search_mark);

  if (!x)
    return 0;

  if (!x->extension_object)
    x->extension_object = TRT_SEND ((reference_imp), _mr_c_tom_Extension,
				    SEL (r_new_p), x);
  return x->extension_object;
}

struct trtd_method *
trt_extension_get_imp (struct trtd_extension *x, selector sel)
{
  int i, n;

  if (x->methods)
    for (i = 0, n = x->methods->num_methods; i < n; i++)
      if (trt_selectors_equal (sel, x->methods->methods[i].sel))
	return &x->methods->methods[i];

  return 0;
}

tom_byte
i_tom_Extension_o_implements_s (tom_object self, selector cmd, selector sel)
{
  struct _es_i_tom_Extension *this
    = trt_ext_address (self, _ei_i_tom_Extension);

  return !!trt_extension_get_imp (this->rti, sel);
}

tom_object
i_tom_Extension_r_meta (tom_object self, selector cmd)
{
  struct _es_i_tom_Extension *this
    = trt_ext_address (self, _ei_i_tom_Extension);
  struct trtd_extension *x = this->rti;

  return (void *) x->meta;
}

tom_object
i_tom_Extension_r_name (tom_object self, selector cmd)
{
  struct _es_i_tom_Extension *this
    = trt_ext_address (self, _ei_i_tom_Extension);
  struct trtd_extension *x = this->rti;

  return x->name.len ? byte_string_with_string (x->name.s, x->name.len) : 0;
}

char *
trt_type_name (enum trt_type_encoding type)
{
  switch (type)
    {
    case TRT_TE_VOID: return "void";
    case TRT_TE_BOOLEAN: return "boolean";
    case TRT_TE_BYTE: return "byte";
    case TRT_TE_CHAR: return "char";
    case TRT_TE_INT: return "int";
    case TRT_TE_LONG: return "long";
    case TRT_TE_FLOAT: return "float";
    case TRT_TE_DOUBLE: return "double";
    case TRT_TE_POINTER: return "pointer";
    case TRT_TE_SELECTOR: return "selector";
    case TRT_TE_REFERENCE: return "reference";
    case TRT_TE_DYNAMIC: return "dynamic";
    default:
      break;
    }

  {
    /* XXX This is a leak.  */
    char *s = xmalloc (30);

    sprintf (s, "<unknown type %d>", type);
    return s;
  }
}

void
trt_unrecognized (tom_object self, selector cmd)
{
  tom_object c = TRT_SEND ((reference_imp), _mr_c_tom_SelectorCondition,
			   SEL (r_for_r_class_r_message_r_selector_s), self,
			   c_tom_Conditions_unrecognized_selector,
			   byte_string_with_c_string ("unrecognized selector"),
			   cmd);
  TRT_SEND ((reference_imp), c, SEL (v_raise));
}

void
vquit (char *fmt, va_list ap)
{
  vfprintf (stderr, fmt, ap);
  fprintf (stderr, "\n");
  fflush (stderr);

  /* This _must_ abort.  Raising a condition isn't good enough.  */
  abort ();
}

void
vfatal (char *fmt, va_list ap)
{
  fprintf (stderr, "trt: fatal error: ");
  vquit (fmt, ap);
}

void
fatal (char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  vfatal (fmt, ap);
  va_end (ap);
}

void
unimplemented (char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  fprintf (stderr, "trt: unimplemented feature: ");
  vquit (fmt, ap);
  va_end (ap);
}

static int
actual_main (int argc, char **argv)
{
  i_tom_MutableObjectArray *args;
  int i, r;

  setvbuf (stderr, NULL, _IOLBF, BUFSIZ);

  if (!trt_metas)
    trt_resolve_module (&trt_main_module_info);

  alloc_init ();

  c_tom_Runtime_long_program_name
    = (void *) byte_string_with_c_string (argv[0]);

  /* XXX Removing everything up to, including, the last `/' can be done
     slightly faster than this...  */
  {
    tom_object a = TRT_SEND ((reference_imp), c_tom_Runtime_long_program_name,
  SEL (r_componentsSeparatedBy_c_limit__i_excludeEmpty__o_substringSelector__s),
			     (tom_char) '/', -1, 0, SEL (r_substring__ii_));
    /* GGG This expression can not be passed instead of the L in the
       TRT_SEND below, since cpp does not expand it there...  */
    int l = TRT_SEND (, a, SEL (i_length));

    c_tom_Runtime_program_name
      = TRT_SEND ((reference_imp), a, SEL (r_at_i), l - 1);
  }

  args = (void *) TRT_SEND (, _mr_c_tom_MutableObjectArray, SEL (r_alloc));
  for (i = 1; i < argc; i++)
    TRT_SEND (, args, SEL (v_add_r), byte_string_with_c_string (argv[i]));

  c_tom_Runtime_all_arguments = TRT_SEND ((reference_imp),
					  _mr_c_tom_ObjectArray,
					  SEL (r_withEnumerable_r), args);

  init_streams ();

  TRT_SEND (, _mr_c_tom_Runtime, SEL (v_preload_r), args);

  trt_execute_load_imps (&trt_main_module_info, args);

  c_tom_Runtime_arguments = TRT_SEND ((reference_imp), _mr_c_tom_ObjectArray,
				      SEL (r_withEnumerable_r), args);

  r = TRT_SEND (, _mr_c_tom_Runtime, SEL (i_start__rs__arguments_r),
		*trt_main_module_info.main_class,
		trt_main_module_info.main_selector, args);

  tini_streams ();

  return r;
}

int
main (int argc, char **argv)
{
#if !STACK_REF_STRUCT
  bottom_of_stack = (void *) &argc;
#endif

#ifdef PATCH_387_CW
  {
    volatile short cw;

    __asm__ volatile ("fstcw (%0)" : : "g" (&cw));
    cw |= 1;			/* Mask the `invalid' exception.  */
    __asm__ volatile ("fldcw (%0)" : : "g" (&cw));
  }
#endif

  return actual_main (argc, argv);
}
