/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: perform.c,v 1.16 1996/08/21 10:23:50 tiggr Exp $  */

#include "trt.h"
#include <tom/tom-r.h>
#include <stdarg.h>
#include <stdio.h>

void
i_tom_Invocation_v_dealloc (tom_object self, selector cmd)
{
  unimplemented ("[Invocation dealloc]");
}

tom_object
i_tom_Invocation_r_fire (tom_object self, selector cmd)
{
  unimplemented ("[Invocation fire]");

  return 0;
}

void
i_tom_InvocationResult_v_dealloc (tom_object self, selector cmd)
{
  unimplemented ("[InvocationResult dealloc]");
}

static void
perform_args (void (*imp) (void), tom_object object, selector sel,
	      selector cmd, int skip_in, va_list ap)
{
  int args_size = 2 * sizeof (void *) + APPLY_ARGS_REG_SIZE;
  union apply_args *args = alloca (args_size);
  void *result;
  int i;

  /* The expected return types must match, otherwise the caller expects
     something the callee can't give him.  */
  if (!trt_selector_args_match (cmd->out, sel->out))
    fatal ("perform return value mismatch: %s, %s",
	   cmd->name.s, sel->name.s);

  APPLY_ARGS_START (sel);
  args->stack = alloca (APPLY_ARGS_STACK_SIZE (sel));
  APPLY_ARGS_EMIT_REFERENCE (object);
  APPLY_ARGS_EMIT_SELECTOR (sel);

  if (sel->in->num != cmd->in->num - skip_in)
    fatal ("perform argument count mismatch: %s, %s",
	   cmd->name.s, sel->name.s);

  for (i = 0; i < sel->in->num; i++)
    {
      if (sel->in->args[i] != cmd->in->args[i + skip_in])
	fatal ("perform argument type mismatch at %d: %s, %s",
	       i, cmd->name.s, sel->name.s);

      switch (sel->in->args[i])
	{
	case TRT_TE_BOOLEAN:
	case TRT_TE_BYTE:
	  APPLY_ARGS_EMIT_BYTE (va_arg (ap, tom_byte));
	  break;

	case TRT_TE_CHAR:
	  APPLY_ARGS_EMIT_CHAR (va_arg (ap, tom_char));
	  break;

	case TRT_TE_INT:
	  APPLY_ARGS_EMIT_INT (va_arg (ap, tom_int));
	  break;

	case TRT_TE_LONG:
	  APPLY_ARGS_EMIT_LONG (va_arg (ap, tom_long));
	  break;

	case TRT_TE_FLOAT:
	  APPLY_ARGS_EMIT_FLOAT (va_arg (ap, tom_float));
	  break;

	case TRT_TE_DOUBLE:
	  APPLY_ARGS_EMIT_DOUBLE (va_arg (ap, tom_double));
	  break;

	case TRT_TE_POINTER:
	case TRT_TE_SELECTOR:
	case TRT_TE_REFERENCE:
	  APPLY_ARGS_EMIT_REFERENCE (va_arg (ap, void *));
	  break;

	case TRT_TE_VOID:
	default:
	  /* This can not happen?  */
	  fatal ("argument type %d at %d to selector %s",
		 sel->in->args[i], i, sel->name.s);
	  break;
	}
    }

  if (sel->out->num > 1)
    {
      for (i = 1; i < cmd->out->num; i++)
	switch (sel->out->args[i])
	  {
	  case TRT_TE_VOID:
	    /* This can not happen?  */
	    fatal ("void return value type at %d to selector %s",
		   i, sel->name.s);
	    break;

	  default:
	    APPLY_ARGS_EMIT_POINTER (va_arg (ap, void **));
	  }
    }

  va_end (ap);

  APPLY_ARGS_COMPLETE ();

  result = __builtin_apply ((void (*) ()) imp, args, APPLY_ARGS_ACTUAL_SIZE ());

  APPLY_ARGS_END ();

  __builtin_return (result);
}

static void
perform_objects (void (*imp) (void), tom_object object, selector sel,
		 selector cmd, tom_object argv, va_list ap)
{
  int args_size = 2 * sizeof (void *) + APPLY_ARGS_REG_SIZE;
  union apply_args *args = alloca (args_size);
  void *result;
  int i, argc;

  /* The expected return types must match, otherwise the caller expects
     something the callee can't give him.  */
  if (!trt_selector_args_match (cmd->out, sel->out))
    fatal ("perform return value mismatch: %s, %s", cmd->name.s, sel->name.s);

  APPLY_ARGS_START (sel);
  args->stack = alloca (APPLY_ARGS_STACK_SIZE (sel));
  APPLY_ARGS_EMIT_REFERENCE (object);
  APPLY_ARGS_EMIT_SELECTOR (sel);

  argc = argv ? TRT_SEND ((int_imp), argv, SEL (i_length)) : 0;
  if (argc != sel->in->num)
    {
      /* Wrong number of arguments.  */
      fatal ("wrong number of arguments %d to selector %s (should be %d)",
	     (int) argc, sel->name.s, sel->in->num);
    }

  if (sel->in->num)
    {
      for (i = 0; i < sel->in->num; i++)
	{
	  switch (sel->in->args[i])
	    {
	    case TRT_TE_BOOLEAN:
	    case TRT_TE_BYTE:
	      APPLY_ARGS_EMIT_BYTE (TRT_SEND ((byte_imp), argv,
					      SEL (b_at_i), i));
	      break;

	    case TRT_TE_CHAR:
	      APPLY_ARGS_EMIT_CHAR (TRT_SEND ((char_imp), argv,
					      SEL (c_at_i), i));
	      break;

	    case TRT_TE_INT:
	      APPLY_ARGS_EMIT_INT (TRT_SEND ((int_imp), argv,
					     SEL (i_at_i), i));
	      break;

	    case TRT_TE_LONG:
	      APPLY_ARGS_EMIT_LONG (TRT_SEND ((long_imp), argv,
					      SEL (l_at_i), i));
	      break;

	    case TRT_TE_FLOAT:
	      APPLY_ARGS_EMIT_FLOAT (TRT_SEND ((float_imp), argv,
					       SEL (f_at_i), i));
	      break;

	    case TRT_TE_DOUBLE:
	      APPLY_ARGS_EMIT_DOUBLE (TRT_SEND ((double_imp), argv,
						SEL (d_at_i), i));
	      break;

	    case TRT_TE_REFERENCE:
	      APPLY_ARGS_EMIT_REFERENCE (TRT_SEND ((reference_imp), argv,
						   SEL (r_at_i), i));
	      break;

	    case TRT_TE_VOID:
	    case TRT_TE_POINTER:
	    case TRT_TE_SELECTOR:
	    default:
	      /* This can not happen?  */
	      fatal ("argument type %d at %d to selector %s",
		     sel->in->args[i], i, sel->name.s);
	      break;
	    }
	}
    }

  if (sel->out->num > 1)
    {
      for (i = 1; i < cmd->out->num; i++)
	switch (sel->out->args[i])
	  {
	  case TRT_TE_VOID:
	    /* This can not happen?  */
	    fatal ("void return value type at %d to selector %s",
		   i, sel->name.s);
	    break;

	  default:
	    APPLY_ARGS_EMIT_POINTER (va_arg (ap, void **));
	  }
    }

  va_end (ap);

  APPLY_ARGS_COMPLETE ();

  result = __builtin_apply ((void (*) ()) imp, args, APPLY_ARGS_ACTUAL_SIZE ());

  APPLY_ARGS_END ();

  __builtin_return (result);
}

void
trt_forward (tom_object self, selector cmd, ...)
{
  int args_size = 2 * sizeof (void *) + APPLY_ARGS_REG_SIZE;
  union apply_args *args = alloca (args_size);
  tom_object delegate = self, fwd_args = 0;
  void_imp fwd_imp = 0;
  selector fwd_sel;
  void *result;
  va_list ap;
  int i, n;
  char *s;

  delegate = TRT_SEND ((reference_imp), self,
		       SEL (r_forwardDelegate_s), cmd);
  if (delegate != self)
    {
      /* XXX There are two options here.  Either we change the FWD_SEL
         selector of the method to be invoked from `forward::' into the
         proper `perform::' or we construct the method invocation of the
         delegate method directly from our own arguments.  The second case
         is (much) faster than the first.  The first case has the
         disadvantage that a selector might need to be constructed
         especially for this purpose.  Either case is equally flexible.

	 We'll wait until anybody needs it.  */
      unimplemented ("non-self forwardDelegate");
    }

  /* Get frobnicated return type.
     XXX Wed Jun 26 01:33:57 1996, tiggr@tricky.es.ele.tue.nl.  */
  for (s = cmd->name.s; *s != '_'; s++);

  n = s - cmd->name.s;
  s = alloca (n + 15);
  memcpy (s, cmd->name.s, n);
  memcpy (s + n, "_forward__s___r", 15);
  fwd_sel = trt_selector_named (s, n + 15);

  if (fwd_sel)
    fwd_imp = (void_imp) trt_lookup ((void *) delegate, fwd_sel);

  if (!fwd_sel || (void (*) ()) fwd_imp == (void (*) ()) trt_forward)
    trt_unrecognized (self, cmd);

  ASSERT (cmd->out->num == fwd_sel->out->num);

  APPLY_ARGS_START (fwd_sel);
  args->stack = alloca (APPLY_ARGS_STACK_SIZE (fwd_sel));

  APPLY_ARGS_EMIT_REFERENCE (delegate);
  APPLY_ARGS_EMIT_SELECTOR (fwd_sel);
  APPLY_ARGS_EMIT_SELECTOR (cmd);

  va_start (ap, cmd);

  if (cmd->in->num)
    {
      fwd_args
	= (void *) TRT_SEND ((reference_imp), _mr_c_tom_MutableObjectArray,
			     SEL (r_alloc));

      for (i = 0; i < cmd->in->num; i++)
	{
	  tom_object o;

	  switch (cmd->in->args[i])
	    {
	    case TRT_TE_VOID:
	    default:
	      /* This can not happen?  */
	      fatal ("void argument type at %d to selector %s", i, cmd->name.s);
	      o = 0;
	      break;

	    case TRT_TE_POINTER:
	    case TRT_TE_SELECTOR:
	      /* XXX */
	      fatal ("pointer/selector argument at %d to trt_forward for %s",
		     i, cmd->name.s);
	      o = 0;
	      break;

	    case TRT_TE_BOOLEAN:
	    case TRT_TE_BYTE:
	      o = TRT_SEND ((reference_imp), _mr_c_tom_ByteNumber,
			    SEL (r_with_b), va_arg (ap, tom_byte));
	      break;

	    case TRT_TE_CHAR:
	      o = TRT_SEND ((reference_imp), _mr_c_tom_CharNumber,
			    SEL (r_with_c), va_arg (ap, tom_char));
	      break;

	    case TRT_TE_INT:
	      o = TRT_SEND ((reference_imp), _mr_c_tom_IntNumber,
			    SEL (r_with_b), va_arg (ap, tom_int));
	      break;

	    case TRT_TE_LONG:
	      o = TRT_SEND ((reference_imp), _mr_c_tom_LongNumber,
			    SEL (r_with_l), va_arg (ap, tom_long));
	      break;

	    case TRT_TE_FLOAT:
	      o = TRT_SEND ((reference_imp), _mr_c_tom_FloatNumber,
			    SEL (r_with_f), va_arg (ap, tom_float));
	      break;

	    case TRT_TE_DOUBLE:
	      o = TRT_SEND ((reference_imp), _mr_c_tom_DoubleNumber,
			    SEL (r_with_d), va_arg (ap, tom_double));
	      break;

	    case TRT_TE_REFERENCE:
	      o = va_arg (ap, tom_object);
	      break;
	    }

	  TRT_SEND (, fwd_args, SEL (v_add_r), o);
	}
    }

  APPLY_ARGS_EMIT_REFERENCE (fwd_args);

  if (cmd->out->num > 1)
    for (i = 1; i < cmd->out->num; i++)
      APPLY_ARGS_EMIT_POINTER (va_arg (ap, void **));

  va_end (ap);

  APPLY_ARGS_COMPLETE ();

  result = __builtin_apply ((void (*) ()) fwd_imp, args,
			    APPLY_ARGS_ACTUAL_SIZE ());

  APPLY_ARGS_END ();

  __builtin_return (result);
}

static void_imp
get_imp (struct trtd_extension *x, selector sel)
{
  struct trtd_method *m = trt_extension_get_imp (x, sel);

  if (!m)
    {
      tom_object c = TRT_SEND
	((reference_imp), _mr_c_tom_SelectorCondition,
	 SEL (r_for_r_class_r_message_r_selector_s), x->extension_object,
	 c_tom_Conditions_unrecognized_selector,
	 byte_string_with_c_string ("method not implemented by this extension"),
	 sel);
      TRT_SEND ((reference_imp), c, SEL (v_raise));
    }

  return m->imp;
}

void
i_tom_All_x_perform_s_with_x
  (tom_object self, selector cmd, selector sel, ...)
{
  void_imp imp = trt_lookup ((struct trt_instance *) self, sel);
  va_list ap;

  va_start (ap, sel);
  perform_args (imp, self, sel, cmd, 1, ap);
}

void
i_tom_All_x_perform_s___r
  (tom_object self, selector cmd, selector sel, tom_object argv, ...)
{
  void_imp imp = trt_lookup ((struct trt_instance *) self, sel);
  va_list ap;

  va_start (ap, argv);
  perform_objects (imp, self, sel, cmd, argv, ap);
}

void
i_tom_Extension_x_perform_s_on_r_with_x
  (tom_object self, selector cmd,
   selector sel, tom_object object, ...)
{
  struct _es_i_tom_Extension *this
    = trt_ext_address (self, _ei_i_tom_Extension);
  void (*m) (void) = get_imp (this->rti, sel);
  va_list ap;

  va_start (ap, object);

  /* XXX Invoke `perform_args' as the last thing in this function (and
     others as well; there is only this one comment about that usage).
     Hopefully, the return value set by `perform_args' will survive
     through this function's exit to reach our caller.  */
  perform_args (m, object, sel, cmd, 2, ap);
}

void
i_tom_Extension_x_perform_s_on_r___r
  (tom_object self, selector cmd,
   selector sel, tom_object object, tom_object args, ...)
{
  struct _es_i_tom_Extension *this
    = trt_ext_address (self, _ei_i_tom_Extension);
  void (*m) (void) = get_imp (this->rti, sel);
  va_list ap;

  va_start (ap, args);
  perform_objects (m, object, sel, cmd, args, ap);
}
