/* 
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: trt.h,v 1.28 1996/08/20 15:59:39 tiggr Exp $  */

#include "tom/trt.h"
#include "tom/util.h"
#include <tom/tom-r.h>
#include <config/target.h>
#include <string.h>

/* Concatenation macros.  */
#define C2(A, B)  D9 (A, B, , , , , , , )
#define C3(A, B, C)  D9 (A, B, C, , , , , , )
#define C4(A, B, C, D)  D9 (A, B, C, D, , , , , )
#define C5(A, B, C, D, E)  D9 (A, B, C, D, E, , , , )
#define C6(A, B, C, D, E, F)  D9 (A, B, C, D, E, F, , , )
#define C7(A, B, C, D, E, F, G)  D9 (A, B, C, D, E, F, G, , )
#define C8(A, B, C, D, E, F, G, H)  D9 (A, B, C, D, E, F, G, H, )
#define C9(A, B, C, D, E, F, G, H, I)  D9 (A, B, C, D, E, F, G, H, I)
#define D9(A, B, C, D, E, F, G, H, I)  A ## B ## C ## D ## E ## F ## G ## H ## I

void init_streams (void);
void tini_streams (void);
void alloc_init (void);

#if !STACK_REF_STRUCT
extern void *bottom_of_stack;
#endif

/* Obscure stuff for GNU CC's apply.  */
union apply_args
{
  void **stack;
  char reg[0];
};

typedef BUILTIN_RETURN_TYPE builtin_return_type;

#define NUM_SELECTORS  (trt_selectors->st.num)
#define NUM_BUCKETS  ((NUM_SELECTORS + TRT_BUCKET_SIZE - 1) / TRT_BUCKET_SIZE)

void trt_forward (tom_object self, selector cmd, ...);

/* Declarations for the runtime structures.  */

/* The default dispatch table.  */
extern struct trt_method_dispatch_table *dtable_default;

/* All known selectors.  */
extern struct trtd_selectors *trt_selectors;

/* The known metas.  The metas include only the class objects (the meta
   class objects are available through these.  */
extern struct trt_metas *trt_metas;

extern tom_object trt_all_classes;

/* The module information of the program.  The reason for the main program's
   module info to have another name than the usual (dynamic loading) name is
   because rld(3) on the NeXT is braindead.  */
extern struct trtd_module trt_main_module_info;

/* The number of extensions id's in use.  */
extern int trt_num_eids;

/* The mark used when traversing the inheritance graph.  This is
   increamented for each following traversal.  */
extern int trt_search_mark;

/* The recorded dynamic selectors.  */
extern struct trt_dynamic_selectors *trt_dynamic_selectors;

/* Grow the stalloc governed memory at *PST to be able to accomodate NUM
   entries of ELT_SIZE bytes each.  The size of the struct with zero
   elements is the ZERO_SIZE.  Iff GRANULARITY, it indicates the increment
   in elements allowed for the capacity.  If it is 0, the capacity starts
   (from 0) at 4 and is doubled in size at other times.  */
void *stalloc (struct stalloc **pst, int zero_size, int elt_size,
	       int num, int granularity);

/* Return whether the SUPER is a supermeta of the SUB meta.  The first
   function searches from the subclass towards the superclass; the second
   one (with reversed arguments!) searches towards the subclass.  */
tom_byte trt_supermeta_star (struct trt_class *sub, struct trt_class *super,
			     int mark);
tom_byte trt_submeta_star (struct trt_class *super, struct trt_class *sub,
			   int mark);

struct trtd_method *trt_extension_get_imp (struct trtd_extension *x,
					   selector sel);

/* Functions exported by array.  */

/* Return a new bytestring which can hold LEN bytes.  Return in THIS the
   pointer to the array extension of the string.  */
i_tom_ByteString new_byte_string (struct _es_i_tom_Array **this, tom_int len);

/* Adjust the capacity of the mutable array defined by A and M which holds
   ELT_SIZE elements, so it can accomodate an extra NUM elements.  */
void trt_array_resize (struct _es_i_tom_Array *a,
		       struct _es_i_tom_MutableArray *m,
		       int elt_size, int num);

/* Functions exported by dload.  */

void trt_resolve_module (struct trtd_module *mi);

void trt_execute_load_imps (struct trtd_module *mi, void *args);

/* List functionality.  */

struct cons
{
  void *car, *cdr;
};

static __inline__ struct cons *
cons (void *a, void *d)
{
  struct cons *c = xmalloc (sizeof (*c));
  c->car = a;
  c->cdr = d;
  return c;
}

/* Hashtable functionality.  */

struct bucket
{
  void *cdr, *key, *value;
};

struct hashtable
{
  int (*hash) (void *);
  int (*equal) (void *, void *);

  /* The number of buckets (power of 2), and the number of contained
     elements.  */
  int num_buckets, num_elts;

  struct bucket **buckets;
};

/* Pointer and string hashtable functions.  */
int hash_hash_pointer (void *elt);
int hash_hash_string (void *elt);
int hash_equal_pointer (void *a, void *b);
int hash_equal_string (void *a, void *b);
int hash_hash_selector (selector sel);

struct bucket *hash_add (struct hashtable *ht, void *key, void *value);
void *hash_find (struct hashtable *ht, void *key);
